/* Copyright 1993 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Routines For Invariant Checking */
/* By Alan J. Hu */
/* January 4, 1993 */

/* ************************ CHANGE LOG ************************** */
/*
24 August 1993	-- Started adding CTL model checking with implicitly
			conjoined invariants.  (Maybe only do ACTL first.)
8 September 1993 -- Added inv_image_ic(), which computes the backward image
			on an implicit conjunction, but uses ORs in the
			nextstate relation to further partition the result
			implicit conjunction (computes each rule individually)
28 October 1993 -- Started adding Nicolas's approximation stuff.
26 November 1993 -- Added more Nicolas stuff.  Also modified inv_image to
			modularize out the BDD back image part (so I could
			call it from elsewhere).
17 March 1994 -- Been messing around with bdd_cofactor vs. bdd_reduce.
26 March 1994 -- Adding code to handle check_with_support.
5 May 1994 -- Adding code to propagate don't-care info while computing BackImage
7 September 1994 -- Moved CTL stuff to file everctl.c.
		    Moved basic conjunct and disjunct list stuff to everlist.c.
		    Moved image computation stuff to everimage.c.
10 September 1995 -- Disabled call to select_small_frontier, could result
			in non-monotonic sequence.
24 September 1995 -- Reenabled call to select_small_frontier.  I had
			misinterpreted my code.
*/
/* ************************************************************** */

#include "ever.h"
#include "everlist.h"	/* Stuff for conjunct and disjunct lists. */
#include "everimage.h"	/* Stuff for image computation. */
#include "bddajh.h"
#include "y.tab.h"	/* Defines constants for the tokens. */


/* These are now forward declarations that could probably be eliminated
	now that I've moved these functions to this file. */

bdd bdd_print_example(bdd cond);	/* Helper function in evermain.c */

bdd compute_forward_image(bdd x, uneval_prop next, int force);
	/* Helper function in evermain.c */
	/* Computes image next(x).  If you modify this, be sure to
	   update compute_backward_image as well. */

bdd compute_backward_image(bdd y, uneval_prop next);
	/* Computes image pred(y).  If you modify this, be sure to
	   update compute_forward_image as well. */

bdd simplify_bdd(bdd f, bdd c)
	/* Tries to find a small bdd that's equal to f for all valuations
	   for which c is true.  (c is the care set.) */
	/* This causes a time penalty, but makes it easier for me to
		improve this. */
{
	return bdd_reduce(gbm,f,c);
	/*return bdd_cofactor(gbm,f,c);*/
}


/* This is a kludgey way to do this, but it's easy. */
/*#define MAX_DEPTH 128*/
#define MAX_DEPTH 511
conjunct_list	I[MAX_DEPTH];
disjunct_list	R[MAX_DEPTH];

/* Various tests for convergence. */

int convergence_1(conjunct_list a, conjunct_list b)
	/* This one does a component by component equality check. */
	/* Sufficient but not necessary. */
{
	return conjunct_list_component_equality(a,b);
}


int convergence_2(conjunct_list a, conjunct_list b)
	/* This one is necessary and sufficient.  It works by checking
		implication in both directions. */
{
    int i;

    fprintf(stderr,"(convergence_2)"); fflush(stderr);
    /* Check that a implies b. */
    /* a implies (b1&b2) if and only if (a implies b1) and (a implies b2) */
    for (i=0; i < b->size; i++) {
	if (!conjunct_list_implies_bdd(a,b->conjunct[i])) return FALSE;
    }

    /* Check that b implies a. */
    for (i=0; i < a->size; i++) {
	if (!conjunct_list_implies_bdd(b,a->conjunct[i])) return FALSE;
    }

    return TRUE;
}

int convergence_2_one_sided(conjunct_list a, conjunct_list b)
	/* Check whether b implies a. */
	/* This is just half of convergence_2, since the fixpoint iterations
		are monotonic, so I need only check implication. */
	/* This one is necessary and sufficient.*/
{
    int i;

    fprintf(stderr,"(convergence_2_one_sided)"); fflush(stderr);

    /* Check that b implies a. */
    /*
    for (i=0; i < a->size; i++) {
	if (!conjunct_list_implies_bdd(b,a->conjunct[i])) return FALSE;
    }

    return TRUE;
    */

    return check_conjunct_list_implication(b,a);
}



/* Various means of conjunct and simplify. */

conjunct_list c_and_s_1(conjunct_list a, conjunct_list b)
	/* This one leaves out[0..n-2]=a[0..n-2], and puts the rest of the
	   stuff, after simplification, into out[n-1]. */
{
    int i, j;
    conjunct_list new, simp_b;
    bdd temp1;

    new = copy_conjunct_list(a);
    simp_b = copy_conjunct_list(b);

    fprintf(stderr,"Simplify pass 1...\n"); /* Simplify each b by all a */
    for (i=0; i < simp_b->size; i++) {
	for (j=0; j < a->size; j++) {
	    temp1 = simplify_bdd(simp_b->conjunct[i], a->conjunct[j]);

	    fprintf(stderr,"Simplify %d by %d -- ",i,j);
	    fprintf(stderr,"Before %ld ",bdd_size(gbm,simp_b->conjunct[i],1));
	    fprintf(stderr,"After %ld\n",bdd_size(gbm,temp1,1));

	    bdd_free(gbm,simp_b->conjunct[i]);
	    simp_b->conjunct[i] = temp1;
	}
    }

    fprintf(stderr,"Conjuncting...\n"); /* Conjunct everything together. */
    for (i=0; i < simp_b->size; i++) {
	temp1 = bdd_and(gbm, new->conjunct[new->size-1], simp_b->conjunct[i]);

	fprintf(stderr,"Size %ld\n",bdd_size(gbm,temp1,1));

	bdd_free(gbm, new->conjunct[new->size-1]);
	new->conjunct[new->size-1] = temp1;
    }

    free_conjunct_list(simp_b);

    fprintf(stderr,"Simplify pass 2...\n"); /* Simplify last a by other a */
    for (i=0; i < new->size-1; i++) {
	temp1 = simplify_bdd(new->conjunct[new->size-1], new->conjunct[i]);

	fprintf(stderr,"Size %ld\n",bdd_size(gbm,temp1,1));

	bdd_free(gbm, new->conjunct[new->size-1]);
	new->conjunct[new->size-1] = temp1;
    }

    return new;
} /* c_and_s_1 */

conjunct_list c_and_s_2(conjunct_list a, conjunct_list b)
	/* Assumes a and b are same length.  Tries to make output
		conjunct list have same value as a&b, but with the
		same number of conjuncts as a. */
	/* This one simplifies each b by all a's, then does a
	   component-wise conjunction out[i] = a[i]&b[i]. */
{
    int i, j;
    conjunct_list new, simp_b;
    bdd temp1;

    if (a->size != b->size) {
	fprintf(stderr,"Error in c_and_s_2:  %d != %d\n",a->size,b->size);
	return NULL;	/* Bail out. */
    }

    new = new_conjunct_list(a->size);
    simp_b = copy_conjunct_list(b);

    fprintf(stderr,"Simplify pass 1...\n"); /* Simplify each b by all a */
    for (i=0; i < simp_b->size; i++) {
	for (j=0; j < a->size; j++) {
	    temp1 = simplify_bdd(simp_b->conjunct[i], a->conjunct[j]);

	    fprintf(stderr,"Simplify %d by %d -- ",i,j);
	    fprintf(stderr,"Before %ld ",bdd_size(gbm,simp_b->conjunct[i],1));
	    fprintf(stderr,"After %ld\n",bdd_size(gbm,temp1,1));

	    bdd_free(gbm,simp_b->conjunct[i]);
	    simp_b->conjunct[i] = temp1;
	}
    }

    fprintf(stderr,"Conjuncting...\n"); /* Conjunct everything together. */
    for (i=0; i < simp_b->size; i++) {
	new->conjunct[i] = bdd_and(gbm, a->conjunct[i], simp_b->conjunct[i]);

	fprintf(stderr,"Size of %d is %ld\n",i,bdd_size(gbm,new->conjunct[i],1));
    }

    free_conjunct_list(simp_b);

    return new;
} /* c_and_s_2 */


conjunct_list c_and_s_3(conjunct_list a, conjunct_list b)
	/* Assumes a and b are same length.  Tries to make output
		conjunct list have same value as a&b, but with the
		same number of conjuncts as a. */
	/* This one does the component-wise conjunction
	   out[i] = a[i]&b[i] first, then does a left-to-right
	   simplification pass. */
{
    int i, j;
    conjunct_list new;
    bdd temp1;

    if (a->size != b->size) {
	fprintf(stderr,"Error in c_and_s_3:  %d != %d\n",a->size,b->size);
	return NULL;	/* Bail out. */
    }

    new = new_conjunct_list(a->size);

    fprintf(stderr,"Conjuncting...\n"); /* Conjunct everything together. */
    for (i=0; i < b->size; i++) {
	new->conjunct[i] = bdd_and(gbm, a->conjunct[i], b->conjunct[i]);

	/*fprintf(stderr,"Size of %d is %ld\n",i,bdd_size(gbm,new->conjunct[i],1));*/
    }

    fprintf(stderr,"Simplify pass 2...\n"); /* Simplify last a by other a */
    for (i=0; i < new->size-1; i++) {
	for (j=i+1; j < new->size; j++) {
	    temp1 = simplify_bdd(new->conjunct[j], new->conjunct[i]);

	    /*fprintf(stderr,"Simplify %d by %d Size %ld\n",j,i,bdd_size(gbm,temp1,1));*/

	    bdd_free(gbm, new->conjunct[j]);
	    new->conjunct[j] = temp1;
	}
    }

    return new;
} /* c_and_s_3 */



conjunct_list evaluate_some_conjunctions(conjunct_list a)
	/* Helper function for c_and_s_4. */
	/* Returns a shorter conjunct list that's equal to a. */
{
#define GROW_THRESHOLD 1.0
    int i,j;
    int besti, bestj;
    double ratio,bestratio;
    bdd size_temp[3];
    int actual_size;
    bdd *pair_table;
    conjunct_list out;

    /* Initialize pairwise conjunction table. */
    fprintf(stderr,"Building %d x %d conjunction table...\n",a->size,a->size);
    actual_size = a->size;
    pair_table = (bdd *)calloc(actual_size*actual_size,sizeof(bdd));
    for (i=0; i<actual_size; i++)
	pair_table[i*(a->size)+i] = bdd_identity(gbm,a->conjunct[i]);
    for (i=0; i<actual_size; i++)
	for (j=i+1; j<actual_size; j++)
	    pair_table[i*(a->size)+j] = bdd_and(gbm,a->conjunct[i],a->conjunct[j]);

    /* Combine Terms Until No More Improvement. */
    size_temp[2] = NULL;
    while (1) {
	/* Find best pair to combine. */
	bestratio = GROW_THRESHOLD+1;
	for (i=0; i<actual_size; i++)
	    for (j=i+1; j<actual_size; j++) {
		/* I should be pre-computing the sizes!  AJH 8/23/93 */
		size_temp[0] = pair_table[i*(a->size)+i];
		size_temp[1] = pair_table[j*(a->size)+j];
		ratio = (double)bdd_size(gbm,pair_table[i*(a->size)+j],1) /
				bdd_size_multiple(gbm,size_temp,1);
		if (ratio<bestratio) {
		    bestratio = ratio;
		    besti = i;
		    bestj = j;
		}
	    }
	fprintf(stderr,"Best ratio is %lg combining %d and %d.  Array size = %d.\n",bestratio,besti,bestj,actual_size);

	if (bestratio > GROW_THRESHOLD) break;	/* Loop Exit */

	/* Update the pairwise conjunction table. */
	/* Free the rows and columns of the entries to be combined. */
	for (i=0; i<=besti; i++)
	    bdd_free(gbm,pair_table[i*(a->size)+besti]);
	for (j=besti+1; j<actual_size; j++)
	    if (j!=bestj) bdd_free(gbm,pair_table[besti*(a->size)+j]);
	for (i=0; i<=bestj; i++)
	    if (i!=besti) bdd_free(gbm,pair_table[i*(a->size)+bestj]);
	for (j=bestj+1; j<actual_size; j++)
	    bdd_free(gbm,pair_table[bestj*(a->size)+j]);
	/* Compute the new entries related to besti */
	pair_table[besti*(a->size)+besti] = pair_table[besti*(a->size)+bestj];
	for (i=0; i<besti; i++)
	    pair_table[i*(a->size)+besti] = bdd_and(gbm,
						pair_table[i*(a->size)+i],
						pair_table[besti*(a->size)+besti]);
	for (j=besti+1; j<actual_size; j++)
	    if (j!=bestj)
		pair_table[besti*(a->size)+j] = bdd_and(gbm,
						    pair_table[j*(a->size)+j],
						    pair_table[besti*(a->size)+besti]);
	/* Delete the row and column related to bestj. */
	for (i=0; i<actual_size; i++)
	    for (j=bestj+1; j<actual_size; j++)
		if (i<=bestj) {
		    /* Only need to slide early rows to the left. */
		    pair_table[i*(a->size)+j-1] = pair_table[i*(a->size)+j];
		} else {
		    /* Rows below bestj need to be slid up as well. */
		    pair_table[(i-1)*(a->size)+j-1] = pair_table[i*(a->size)+j];
		}
	actual_size--;
    }

    /* Create the output conjunct_list. */
    out = new_conjunct_list(actual_size);
    for (i=0; i<actual_size; i++)
	out->conjunct[i] = bdd_identity(gbm,pair_table[i*(a->size)+i]);
    
    /* Free the pair table. */
    for (i=0; i<actual_size; i++)
	for (j=i; j<actual_size; j++)
	    bdd_free(gbm,pair_table[i*(a->size)+j]);
    free(pair_table);

    return out;
}

conjunct_list c_and_s_4(conjunct_list a, conjunct_list b)
	/* This one handles any sized a and b, and tries to find a good
	   set of conjunctions to evaluate based on BDD sizes. */
{
    int i, j;
    long sizei, sizej, sizetemp;
    conjunct_list out, c;
    bdd temp1;

    /* Append the two lists. */
    c = concat_conjunct_lists(a,b);

    /* Simplify everything by everything. */
    for (i=0; i < c->size; i++) {
	for (j=0; j < c->size; j++) {
	    if (i!=j) {
		sizei = bdd_size(gbm,c->conjunct[i],1);
		sizej = bdd_size(gbm,c->conjunct[j],1);

		/*fprintf(stderr,"Simplify %d(%ld) by %d(%ld) --",i,sizei,j,sizej);
		fflush(stderr);*/

		if (sizei>=sizej) {
		    temp1 = simplify_bdd(c->conjunct[i],c->conjunct[j]);

		    /*sizetemp = bdd_size(gbm,temp1,1);
		    fprintf(stderr," Result size (%ld).\n",sizetemp); fflush(stderr);*/
		    /*fprintf(stderr," Done.\n"); fflush(stderr);*/
		    bdd_free(gbm,c->conjunct[i]);
		    c->conjunct[i] = temp1;
		} else {
		    /*fprintf(stderr," Skipped.\n"); fflush(stderr);*/
		}
	    }
	}
    }

    /* Remove duplicates, special cases. */

    /* Munch the list down by evaluating conjunctions. */
    /*out = evaluate_some_conjunctions(c);
    free_conjunct_list(c);

    return out;*/

    slow_munch_conjunct_list(c,GROW_THRESHOLD);
    return c;
} /* c_and_s_4 */





/* Helper functions for bdd_check_invariant */

int find_violation(bdd set, conjunct_list inv)
	/* Returns index of invariant violated, or -1 if all satisfied. */
{
    int i;
    bdd temp;

    for (i=0; i < inv->size; i++) {
	/* David Long changed his bdd_implies function so that it's
		weird now.  It returns a bdd that implies (f and not g).
		Therefore, f implies g iff bdd_implies(f,g) returns bdd_zero. */
	temp = bdd_implies(gbm,set,inv->conjunct[i]);
	if (temp != bdd_zero(gbm)) {
	    bdd_free(gbm,temp);
	    return i;	/* Violates ith conjunct. */
	}
	bdd_free(gbm,temp);
    }

    /* No violation */
    return -1;
}

void find_violation_id_ic(disjunct_list a, conjunct_list b,
				int *which_a, int *which_b)
	/* Finds which disjunct in a doesn't imply which conjunct in b.
	   This is a witness that a doesn't imply b.
	   If no such pair is found, then returns -1, and we know that
	   a does imply b. */
{
    int i, j;
    bdd temp;

    for (i=0; i < a->size; i++) {
	for (j=0; j < b->size; j++) {
	    temp = bdd_implies(gbm,a->disjunct[i],b->conjunct[j]);
		/* Returns bdd_zero iff implication holds. */
	    if (temp != bdd_zero(gbm)) {
		/* We've found a violation. */
		*which_a = i;
		*which_b = j;
		return;
	    }
	    bdd_free(gbm,temp);
	}
    }

    /* No violation */
    *which_a = -1;
    *which_b = -1;
}

int converged(conjunct_list a, conjunct_list b)
	/* Returns true if a and b represent same function. */
	/* This function might be only heuristic. */
{
	/*return convergence_2_one_sided(a,b);*/	/* Fancy heuristics */
	return convergence_1(a,b);	/* Old CAV 93 heuristics */
}



conjunct_list conjunct_and_simplify(conjunct_list a, conjunct_list b)
	/* Tries to create a conjunct list with the same value as a&b,
		but with fewer conjuncts. */
{
	return c_and_s_3(a,b);	/* Old CAV 93 heuristics */
	/*return c_and_s_4(a,b);*/	/* For fancy heuristics */
}

conjunct_list eval_invariant(uneval_prop invariant)
	/* Returns a conjunct list corresponding to the invariant. */
{
    /* This is kind of kludgey.  If the top type is AND, I'll make
	a conjunct list.  Otherwise, I'll just evaluate it completely. */
    conjunct_list new;
    bdd temp1;
    int i, j;

    if (invariant->type==AND) {
	/* Conjunctive partition. */
	int length = 0;
	prop_list tp = invariant->prop_list1;

	/* Compute length. */
	while (tp!=NULL) {
	    length++;
	    tp = tp->next;
	}

	/* Evaluate each conjunct. */
	new = new_conjunct_list(length);
	for (length=0, tp=invariant->prop_list1;
		tp!=NULL;
		length++, tp=tp->next) {
	    evaluate_proposition(tp->data,FALSE);
	    new->conjunct[length] = bdd_identity(gbm,tp->data->bdd);
	}

	/* Simplify the invariants! */
	fprintf(stderr,"Simplifying...\n");
	for (i=0; i < new->size-1; i++) {
	    for (j=i+1; j < new->size; j++) {
		temp1 = simplify_bdd(new->conjunct[j], new->conjunct[i]);

		fprintf(stderr,"Simplify %d by %d Size %ld\n",j,i,bdd_size(gbm,temp1,1));

		bdd_free(gbm, new->conjunct[j]);
		new->conjunct[j] = temp1;
	    }
	}
    } else {
	/* Non-partitioned. */
	evaluate_proposition(invariant,FALSE);
	new = new_conjunct_list(1);
	new->conjunct[0] = bdd_identity(gbm, invariant->bdd);
    }

    return new;
}

disjunct_list eval_startstates(uneval_prop startstates)
	/* Returns a disjunct list corresponding to the startstates. */
{
    /* This is kind of kludgey.  If the top type is OR, I'll make
	a disjunct list.  Otherwise, I'll just evaluate it completely. */
    disjunct_list new;
    bdd temp1;
    int i, j;

    if (startstates->type==OR) {
	/* Disjunctive partition. */
	int length = 0;
	prop_list tp = startstates->prop_list1;

	/* Compute length. */
	while (tp!=NULL) {
	    length++;
	    tp = tp->next;
	}

	/* Evaluate each disjunct. */
	new = new_disjunct_list(length);
	for (length=0, tp=startstates->prop_list1;
		tp!=NULL;
		length++, tp=tp->next) {
	    evaluate_proposition(tp->data,FALSE);
	    new->disjunct[length] = bdd_identity(gbm,tp->data->bdd);
	}

	/* I can probably do simplification here, but I don't want to
		think about it right now.  --AJH 12 September 1994 */
    } else {
	/* Non-partitioned. */
	evaluate_proposition(startstates,FALSE);
	new = new_disjunct_list(1);
	new->disjunct[0] = bdd_identity(gbm, startstates->bdd);
    }

    return new;
}


void print_counterexample_trace(bdd start, uneval_prop next, int depth)
{
    int which_one;
    bdd x, P;
    bdd temp1, temp2;

    P = bdd_identity(gbm,start);

    printf("POSSIBLE INITIAL STATE:\n");
    while (depth>=0) {
	which_one = find_violation(P,I[depth]);
	if (which_one<0) {
	    fprintf(stderr,"Internal Error in print_counterexample_trace!  Bailing out.\n");
	    return;
	}

	/* Choose a state that leads to violation. */
	temp1 = bdd_not(gbm,I[depth]->conjunct[which_one]);
	temp2 = bdd_and(gbm,P,temp1);
	bdd_free(gbm,temp1);
	bdd_free(gbm,P);
	x = bdd_print_example(temp2);	/* Chooses the state and prints it. */
	bdd_free(gbm,temp2);

	/* Compute successors of x */
	temp1 = compute_forward_image(x,next,FALSE);
	bdd_free(gbm,x);
	/* Convert phase back to current phase. */
	bdd_assoc(gbm,next_vars_assoc);
	P = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
	bdd_free(gbm,temp1);

	depth--;
	if (depth>=0) printf("CAN GO TO\n");
    }

    printf("VIOLATES INVARIANT #%d.\n",which_one);
    bdd_free(gbm,P);
}




void bdd_check_invariant(bdd start,
			uneval_prop next,
			conjunct_list invariant)
{
    int depth = 0;
    int i, j;
    conjunct_list new_list;

    I[0] = copy_conjunct_list(invariant);

printf("DEBUG a\n");
    do {
/*printf("DEBUG b\n");*/
	print_conjunct_list_size(I[depth],depth); /* For benchmarking only. */
	if (find_violation(start,I[depth])>=0) {
/*printf("DEBUG c\n");*/
	    /* We have a violation.  Print counterexample trace. */
	    print_counterexample_trace(start,next,depth);

	    /* Free intermediate invariants. */
	    for (i=depth; i>=0; i--) {
		free_conjunct_list(I[i]);
	    }
	    return;
	}
/*printf("DEBUG d\n");*/

	new_list = inv_image(next,I[depth]);
	/*new_list = inv_image_ic(next,I[depth]);*/
/*printf("DEBUG e\n");*/
	I[depth+1] = conjunct_and_simplify(I[0],new_list);
/*printf("DEBUG f\n");*/
	free_conjunct_list(new_list);
printf("DEBUG g\n");
	depth++;
    } while (!converged(I[depth],I[depth-1]));
    print_conjunct_list_size(I[depth],depth); /* For benchmarking only. */

    /* Everything's OK. */
    printf("INVARIANT IS CORRECT.\n");

    /* Free intermediate invariants. */
    for (i=depth; i>=0; i--) {
	free_conjunct_list(I[i]);
    }
} /* bdd_check_invariant */



void check_invariant(uneval_prop startstates,
		     uneval_prop nextstate,
		     uneval_prop invariant)
{
    conjunct_list eval_inv;

    printf("Start of check_invariant.\n");
    reset_memory_meter();
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating start condition...");fflush(stderr);
    evaluate_proposition(startstates,FALSE);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating start condition.\n");
    printf("Start condition size = %ld or %ld\n",
			bdd_size(gbm,startstates->bdd,1),
			count_bdd_nodes(startstates));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating invariant...");fflush(stderr);
    eval_inv = eval_invariant(invariant);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating invariant.\n");
    printf("Invariant size = %ld or %ld\n",
			bdd_size(gbm,invariant->bdd,1),
			count_bdd_nodes(invariant));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    if (global_fdv_count>0) {
	/* I don't support FDVs with check_invariant.  Bail out. */
	fprintf(stderr,"Error:  Cannot use dependent variables with checkinvariant.\n");
	return;
    }

    fprintf(stderr,"Starting check...");fflush(stderr);
    bdd_check_invariant(startstates->bdd,
			nextstate,
			eval_inv);
    free_conjunct_list(eval_inv);
    fprintf(stderr,"\n");fflush(stderr);

    printf("End of printtrace.\n");
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Next state relation uses %ld bdd nodes.\n",
					count_bdd_nodes(nextstate));
    printf("\n");
    fflush(stdout);
}




/* KLUDGE */
/* MAX_DEPTH defined earlier. */
bdd Iwa[MAX_DEPTH];
	/* Something very screwy going on.  If I set MAX_DEPTH to
	 some numbers, the bdd result cache blows up more than for
	 other numbers.  e.g. 511 and 512 blow up for pipe_2r.e with
	 2-bit words (using checkinvariant), but 513 gives a very
	 small memory size. */
    /* I with approximation.  Analogous to I[], defined earlier. */

void print_counterexample_with_approx(bdd start,
					uneval_prop next,
					int depth,
					conjunct_list assumptions)
    /* Turns out, the approximations aren't needed at all, since
       they are upper approximations to the reachable state set, so
       they have no effect on the counterexample. */
{
    bdd x, P;
    bdd temp1, temp2;

    P = bdd_identity(gbm,start);

    printf("POSSIBLE INITIAL STATE:\n");
    while (depth>=0) {
	/* Choose a state that leads to violation. */
	temp1 = bdd_not(gbm,Iwa[depth]);
	temp2 = bdd_and(gbm,P,temp1);
	bdd_free(gbm,temp1);
	bdd_free(gbm,P);
	x = bdd_print_example(temp2);	/* Chooses the state and prints it. */
	bdd_free(gbm,temp2);

	/* Compute successors of x */
	temp1 = compute_forward_image(x,next,FALSE);
	bdd_free(gbm,x);
	/* Convert phase back to current phase. */
	bdd_assoc(gbm,next_vars_assoc);
	P = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
	bdd_free(gbm,temp1);

	depth--;
	if (depth>=0) printf("CAN GO TO\n");
    }

    printf("VIOLATES PROPERTY.\n");
    bdd_free(gbm,P);
};


bdd bdd_check_one_with_approx(bdd start,
				uneval_prop next,
				bdd property,
				conjunct_list assumptions)
	/* Uses a backward traversal to check whether property holds
	   for all states reachable from startstates via nextstate.
	   Uses "assumptions" as a list of upper approximations to
	   the set of reachable states (Nicolas's idea).
	   Returns NULL if the property isn't satisfied.
		(Also responsible for printing counterexample trace.)
	   Returns the largest set guaranteed to remain within property
		if the property is satisfied.  (This fixpoint is an
		upper approximation to the reachable states.) */
{
    int depth = 0;
    int i, j;
    bdd new;
    bdd temp;

    Iwa[0] = bdd_identity(gbm,property);

    fprintf(stderr,"Iteration %d, BDD Size %ld\n",
		    depth,bdd_size(gbm,Iwa[depth],1));

    /* Simplify using the assumptions (upper approx of reachable
       states) as care sets. */
    fprintf(stderr,"\tSimplifying: %ld ",bdd_size(gbm,Iwa[depth],1));
    fflush(stderr);
    for (i=0; i < assumptions->size; i++) {
	/*temp = simplify_bdd(Iwa[depth],assumptions->conjunct[i]);*/
	temp = bdd_cofactor(gbm,Iwa[depth],assumptions->conjunct[i]);
	bdd_free(gbm,Iwa[depth]);
	Iwa[depth] = temp;
	fprintf(stderr,"-> %ld ",bdd_size(gbm,Iwa[depth],1));
	fflush(stderr);
    }
    fprintf(stderr,"\n");

    do {
	/* David Long changed his bdd_implies function so that it's
		weird now.  It returns a bdd that implies (f and not g).
		Therefore, f implies g iff bdd_implies(f,g) returns bdd_zero. */
	temp = bdd_implies(gbm,start,Iwa[depth]);
	if (temp != bdd_zero(gbm)) {
	    bdd_free(gbm,temp);

	    /* We have a violation.  Print counterexample trace. */
	    print_counterexample_with_approx(start,next,depth,assumptions);

	    /* Free intermediate invariants. */
	    for (i=depth; i>=0; i--) {
		bdd_free(gbm,Iwa[i]);
	    }
	    return NULL;
	}
	bdd_free(gbm,temp);

	new = inv_image_bdd(next,Iwa[depth]);

	/* Pre-simplify the BackImage using the assumptions
	   (upper approx of reachable states) as care sets. --AJH 12/1/93 */
	fprintf(stderr,"\tSimplifying BackImage: %ld ",bdd_size(gbm,new,1));
	fflush(stderr);
	for (i=0; i < assumptions->size; i++) {
	    temp = simplify_bdd(new,assumptions->conjunct[i]);
	    bdd_free(gbm,new);
	    new = temp;
	    fprintf(stderr,"-> %ld ",bdd_size(gbm,new,1));
	    fflush(stderr);
	}
	fprintf(stderr,"\n");

	depth++;
	Iwa[depth] = bdd_and(gbm,Iwa[0],new);
	bdd_free(gbm,new);

	fprintf(stderr,"Iteration %d, BDD Size %ld\n",
			depth,bdd_size(gbm,Iwa[depth],1));

	/* Simplify using the assumptions (upper approx of reachable
	   states) as care sets. */
	fprintf(stderr,"\tSimplifying: %ld ",bdd_size(gbm,Iwa[depth],1));
	fflush(stderr);
	for (i=0; i < assumptions->size; i++) {
	    /*temp = simplify_bdd(Iwa[depth],assumptions->conjunct[i]);*/
	    temp = bdd_cofactor(gbm,Iwa[depth],assumptions->conjunct[i]);
	    bdd_free(gbm,Iwa[depth]);
	    Iwa[depth] = temp;
	    fprintf(stderr,"-> %ld ",bdd_size(gbm,Iwa[depth],1));
	    fflush(stderr);
	}
	fprintf(stderr,"\n");
	/*bdd_print_profile(gbm,Iwa[depth],bdd_naming_fn_full,NULL,80,stdout);*/
    } while (Iwa[depth] != Iwa[depth-1]);
	/* Modified termination test based on approximations? */

    /* Everything's OK. */
    printf("PROPERTY IS CORRECT.\n");

    temp = bdd_identity(gbm,Iwa[depth]);

    /* Free intermediate invariants. */
    for (i=depth; i>=0; i--) {
	bdd_free(gbm,Iwa[i]);
    }

    return temp;
}

void bdd_check_with_approx(bdd startstates,
			   uneval_prop nextstate,
			   conjunct_list eval_properties,
			   conjunct_list eval_assumptions)
{
    int which_one;
    conjunct_list assume_temp;
    bdd new_I;

    assume_temp = copy_conjunct_list(eval_assumptions);

    for (which_one=0; which_one < eval_properties->size; which_one++) {
	printf("Checking property %d with %d assumptions...\n",
			which_one, assume_temp->size);
	/* bdd_check_one_with_approx is responsible for printing the error
	   trace if it fails and then returning a NULL.  If the check succeeds,
	   it will return the greatest invariant contained in the property. */
	new_I = bdd_check_one_with_approx(startstates,
					nextstate,
					eval_properties->conjunct[which_one],
					assume_temp);
	if (new_I == NULL) {
	    break;	/* We had a violation.  Stop. */
	} else {
	    append_to_conjunct_list(new_I,assume_temp);
	    bdd_free(gbm,new_I);
	}
    }

    free_conjunct_list(assume_temp);
}


void check_with_approx(uneval_prop startstates,
		     uneval_prop nextstate,
		     uneval_prop properties,
		     uneval_prop assumptions)
{
    conjunct_list eval_properties, eval_assumptions;

    printf("Start of check_with_approx.\n");
    reset_memory_meter();
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating start condition...");fflush(stderr);
    evaluate_proposition(startstates,FALSE);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating start condition.\n");
    printf("Start condition size = %ld or %ld\n",
			bdd_size(gbm,startstates->bdd,1),
			count_bdd_nodes(startstates));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating properties...");fflush(stderr);
    eval_properties = eval_invariant(properties);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating properties.\n");
    printf("Size of properties list:\n");
    print_conjunct_list_size(eval_properties,0);
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating assumptions...");fflush(stderr);
    eval_assumptions = eval_invariant(assumptions);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating assumptions.\n");
    printf("Size of assumptions list:\n");
    print_conjunct_list_size(eval_assumptions,0);
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    if (global_fdv_count>0) {
	/* I don't support FDVs with check_with_approx.  Bail out. */
	fprintf(stderr,"Error:  Cannot use dependent variables with checkwithapprox.\n");
	return;
    }

    fprintf(stderr,"Starting check...");fflush(stderr);
    bdd_check_with_approx(startstates->bdd,
			  nextstate,
			  eval_properties,
			  eval_assumptions);
    free_conjunct_list(eval_properties);
    free_conjunct_list(eval_assumptions);
    fprintf(stderr,"\n");fflush(stderr);

    printf("End of check_with_approx.\n");
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Next state relation uses %ld bdd nodes.\n",
					count_bdd_nodes(nextstate));
    printf("\n");
    fflush(stdout);
}



/* Helper functions for check_with_support. */

conjunct_list project_onto_support_sets(conjunct_list x,
					int *assocs,
					int count)
	/* Returns a conjunct list that consists of one conjunct per
		var assoc.  This conjunct is the AND of (each conjunct of
		x existentially quantified by the assoc). */
{
    conjunct_list new;
    int i, j;
    bdd partial, temp;

    new = new_conjunct_list(count);
    for (i=0; i<count; i++) {
	partial = bdd_one(gbm);

	for (j=0; j < x->size; j++) {
	    /* Should I use bdd_rel_prod or do the operations separately? */
	    bdd_assoc(gbm,assocs[i]);
	    temp = bdd_rel_prod(gbm,partial,x->conjunct[j]);
	    bdd_free(gbm,partial);
	    partial = temp;
	}

	new->conjunct[i] = partial;
    }

    return new;
}

void bdd_check_with_support(bdd start,
			    uneval_prop next,
			    conjunct_list invariant,
			    int *support_assocs,
			    int support_count)
{
    int depth = 0;
    int i, j;
    conjunct_list new_list;
    conjunct_list temp_list;
    conjunct_list projected_list;

    I[0] = copy_conjunct_list(invariant);

printf("DEBUG a\n");
    do {
	print_conjunct_list_size(I[depth],depth); /* For benchmarking only. */
	if (find_violation(start,I[depth])>=0) {
	    /* We have a violation.  Print counterexample trace. */
	    print_counterexample_trace(start,next,depth);

	    /* Free intermediate invariants. */
	    for (i=depth; i>=0; i--) {
		free_conjunct_list(I[i]);
	    }
	    return;
	}

	/*new_list = inv_image(next,I[depth]);*/
	new_list = inv_image_ic(next,I[depth]);
	/*new_list = inv_image_care_set(next,I[depth],I[depth]);*/

	/* Project onto the support sets. */
		/* 12 Apr 1994 -- This will do for now.  Long term,
			consider integrating the projection with
			the conjunction and simplification. --AJH */
	/*
	/*projected_list = project_onto_support_sets(new_list,
	/*					    support_assocs,
	/*					    support_count);
	/*temp_list = concat_conjunct_lists(projected_list,new_list);
	/*free_conjunct_list(projected_list);
	/*free_conjunct_list(new_list);

	/*I[depth+1] = conjunct_and_simplify(I[0],temp_list);
	/*free_conjunct_list(temp_list);
	*/

	/* 17 Apr 1994 -- Let's try forcing things into the form:
			projected bdd, ..., projected bdd, remainder bdd
	*/
	/*temp_list = concat_conjunct_lists(I[0],new_list);
	/*free_conjunct_list(new_list);
	/*projected_list = project_onto_support_sets(temp_list,
	/*					    support_assocs,
	/*					    support_count);
	/*append_to_conjunct_list(bdd_one(gbm),projected_list);
	/*    /* Kludge to reuse c_and_s_1. */
	/*I[depth+1] = c_and_s_1(projected_list,temp_list);
	/*    /* Use a specific conjunct_and_simplify. */
	/*free_conjunct_list(projected_list);
	/*free_conjunct_list(temp_list);
	*/

	/* 29 May 1994 -- Munch down the list BEFORE projection, then
		project, then simplify. */
	fast_munch_conjunct_list(new_list);
	medium_munch_conjunct_list(new_list);
	/* 27 Jun 1994 DEBUG -- print out the supports. */
	/*
	{
	    int i;
	    printf("*** *** BEGIN SUPPORTS *** ***\n");
	    for (i=0; i < new_list->size; i++) {
		printf("--- Conjunct %d:\n",i);
		print_bdd_support(stdout,bdd_naming_fn_full,
					new_list->conjunct[i]);
	    }
	    printf("*** *** END SUPPORTS *** ***\n");
	}
	*/
	/*slow_munch_conjunct_list(new_list,1.0);*/
	phased_slow_munch_conjunct_list(new_list,1.0,2);
	projected_list = project_onto_support_sets(new_list,
						   support_assocs,
						   support_count);
	temp_list = concat_conjunct_lists(projected_list,new_list);
	free_conjunct_list(projected_list);
	free_conjunct_list(new_list);

	I[depth+1] = conjunct_and_simplify(I[0],temp_list);
	free_conjunct_list(temp_list);

printf("DEBUG g\n");
	depth++;
	/*
	{
	    int i;
	    printf("*** *** ITERATION %d *** ***\n", depth);
	    for (i=0; i < I[depth]->size; i++) {
		printf("--- Conjunct %d:\n",i);
		bdd_print_bdd(gbm,I[depth]->conjunct[i],
			bdd_naming_fn_full,
			bdd_terminal_id_fn_none,
			NULL,stdout);
	    }
	    printf("*** *** END *** ***\n");
	    fflush(stdout);
	}
	*/
    } while (!converged(I[depth],I[depth-1]));
    print_conjunct_list_size(I[depth],depth); /* For benchmarking only. */

    /* Everything's OK. */
    printf("INVARIANT IS CORRECT.\n");

    /* Free intermediate invariants. */
    for (i=depth; i>=0; i--) {
	free_conjunct_list(I[i]);
    }
} /* bdd_check_with_support */

int assoc_not_in_support_bit_vector(bit_vector x)
	/* Returns a variable association which includes all variables
	   not in the support of any BDD in the bit_vector x. */
{
    bdd *not_in_support;	/* List of BDD vars not in support. */
    int assoc;			/* New association created. */
    int count;
    int var_index;
    int i;

    /* Make sure x is evaluated. */
    if (x->bit_list==NULL) {
	evaluate_bit_vector(x,FALSE);
    }

    /* Allocate a big block of memory to hold all the BDD vars not in
	the support. */
    not_in_support = (bdd *)malloc(((long)global_bdd_count/3+1)* sizeof(bdd));

    /* Step through all vars in system and check whether they're in support. */
    count = 0;
    var_index = 0;
    while (cur_vars[var_index]!=0) {
	for (i=0; i < x->size; i++) {
	    if (bdd_depends_on(gbm,x->bit_list[i],cur_vars[var_index])) break;
	}
	/* Did we fall through the loop or break out? */
	if (i == x->size) {
	    /* Fell through loop.  Variable is not in support. */
	    not_in_support[count++] = cur_vars[var_index];
	}

	var_index++;
    }
    not_in_support[count] = NULL;	/* Terminator */

    /* Build the association. */
    assoc = bdd_new_assoc(gbm,not_in_support,0);

    free(not_in_support);

    return assoc;
}

/* KLUDGE -- This is an array to hold the variable assoc ids for the
   support sets. */
#define MAX_SUPPORT_SET_COUNT 100
int support_set_assocs[MAX_SUPPORT_SET_COUNT];
int support_set_count;

void check_with_support(uneval_prop startstates,
		     uneval_prop nextstate,
		     uneval_prop properties,
		     bit_vector_list support_set_specifiers)
{
    conjunct_list eval_properties;
    bit_vector_list temp_bvl;
    int i;

    printf("Start of check_with_support.\n");
    reset_memory_meter();
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating start condition...");fflush(stderr);
    evaluate_proposition(startstates,FALSE);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating start condition.\n");
    printf("Start condition size = %ld or %ld\n",
			bdd_size(gbm,startstates->bdd,1),
			count_bdd_nodes(startstates));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating properties...");fflush(stderr);
    eval_properties = eval_invariant(properties);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating properties.\n");
    printf("Size of properties list:\n");
    print_conjunct_list_size(eval_properties,0);
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating support set specifiers");fflush(stderr);
    temp_bvl = support_set_specifiers;
    support_set_count = 0;
    while (temp_bvl!=NULL) {
	/* Evaluate the support set. */
	evaluate_bit_vector(temp_bvl->data,FALSE);
	fprintf(stderr,".");fflush(stderr);

	support_set_count++;
	temp_bvl = temp_bvl->next;
    }
    /* Error check because of kludgey handling of array. */
    if (support_set_count > MAX_SUPPORT_SET_COUNT) {
	fprintf(stderr,"\nERROR:  %d support sets greater than limit of %d.\n",
			support_set_count,MAX_SUPPORT_SET_COUNT);
	fprintf(stderr,"\tRecompile everinv.c with larger value of MAX_SUPPORT_SET_COUNT.\n");
	return;
    }
    fprintf(stderr,"\n");fflush(stderr);

    fprintf(stderr,"Building support set variable associations");
    fflush(stderr);
    temp_bvl = support_set_specifiers;
    i = 0;
    while (temp_bvl!=NULL) {
	support_set_assocs[i] = assoc_not_in_support_bit_vector(temp_bvl->data);
	fprintf(stderr,".");fflush(stderr);

	temp_bvl = temp_bvl->next;
	i++;
    }
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done processing support sets.\n");
    /*printf("Size of assumptions list:\n");*/
    /*print_conjunct_list_size(eval_assumptions,0);*/
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    if (global_fdv_count>0) {
	/* I don't support FDVs with check_with_approx.  Bail out. */
	fprintf(stderr,"Error:  Cannot use support sets with checkwithsupport.\n");
	return;
    }

    fprintf(stderr,"Starting check...");fflush(stderr);
    bdd_check_with_support(startstates->bdd,
			  nextstate,
			  eval_properties,
			  support_set_assocs,
			  support_set_count);
    free_conjunct_list(eval_properties);
    /* Free the var_assocs for the support sets. */
    for (i=0; i < support_set_count; i++) {
	bdd_free_assoc(gbm,support_set_assocs[i]);
    }
    fprintf(stderr,"\n");fflush(stderr);

    printf("End of check_with_support.\n");
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Next state relation uses %ld bdd nodes.\n",
					count_bdd_nodes(nextstate));
    printf("\n");
    fflush(stdout);
}




/* Helper functions for printtrace */

bdd bdd_bits_example(bdd *base,
			     string_list labels,
			     int bit_count,
			     int print_prev,
			     int print_cur,
			     int print_next,
			     bdd *cond)
{
    bdd example = bdd_one(gbm);
    bdd temp1, temp2;
    int numerical_value;
    int i;
    int phase;

    for (phase=0; phase<3; phase++)
	if (((phase==0)&&print_prev) ||
	    ((phase==1)&&print_cur) ||
	    ((phase==2)&&print_next)) {

	/* Build the actual example, while modifying the condition. */
	numerical_value = 0;
	for (i=0; i<bit_count; i++) {
	    bdd a_var = base[3*i+phase];

	    temp1 = bdd_compose(gbm,*cond,a_var,bdd_zero(gbm));
	    if (temp1==bdd_zero(gbm)) {
		bdd_free(gbm,temp1);
		temp1 = bdd_compose(gbm,*cond,a_var,bdd_one(gbm));
		if (temp1==bdd_zero(gbm)) {
		    /* We're stuck! */
		    printf("Internal Error in bdd_bits_example.  No example possible.");
		} else {
		    /* Modify the condition. */
		    bdd_free(gbm,*cond);
		    *cond = temp1;
		    /* This bit should be one. */
		    numerical_value += (1<<i);
		    temp2 = bdd_and(gbm,example,a_var);
		    bdd_free(gbm,example);
		    example = temp2;
		}
	    } else {
		/* Modify the condition. */
		bdd_free(gbm,*cond);
		*cond = temp1;
		/* This bit should be zero. */
		temp1 = bdd_not(gbm,a_var);
		temp2 = bdd_and(gbm,example,temp1);
		bdd_free(gbm,temp1);
		bdd_free(gbm,example);
		example = temp2;
	    }
	}

	/* Actually do the output */
	switch (phase) {
	    case 0: printf("^p = "); break;
	    case 1: printf("^c = "); break;
	    case 2: printf("^n = "); break;
	}
	if (labels==NULL) printf("%d\t", numerical_value);
	else {
	    for (i=0; i<numerical_value; i++) labels = labels->next;
	    printf("%s\t", labels->data);
	}

    }	/* for phase */
    printf("\n");

    return example;
}


void n_tabs(int how_many)
{
	int i;

	for (i=0; i<how_many; i++) printf("\t");
}

bdd bdd_partial_example(bdd *base,
				int depth,
				type_desc remaining_type,
				int pre_offset,
				int print_prev,
				int print_cur,
				int print_next,
				bdd *cond)
{
    bdd example = bdd_one(gbm);
    bdd sub_example, temp1;

    if (remaining_type->type==BITS) {
	return bdd_bits_example(&(base[pre_offset]),
				remaining_type->output_labels,
				remaining_type->bits,
				print_prev,
				print_cur,
				print_next,
				cond);	/* Modified! */
    }

    if (remaining_type->type==ARRAY) {
	int i;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	printf("\n");
	for (i=lb; i<=ub; i++) {
	    n_tabs(depth);
	    printf("[%d]",i);
	    sub_example = bdd_partial_example(base,
					      depth+1,
					      element_type,
					      pre_offset+
						(i-lb)*element_type->bdd_count,
					      print_prev,
					      print_cur,
					      print_next,
					      cond);	/* Modified! */
	    temp1 = bdd_and(gbm,example,sub_example);
	    bdd_free(gbm,example);
	    bdd_free(gbm,sub_example);
	    example = temp1;
	}
	return example;
    }

    if (remaining_type->type==RECORD) {
	record_list temp = remaining_type->field_list;
	int offset = 0;

	printf("\n");
	while (temp!=NULL) {
	    n_tabs(depth);
	    printf(".%s",temp->field_name);
	    sub_example = bdd_partial_example(base,
					      depth+1,
					      temp->field_type,
					      pre_offset+offset,
					      print_prev,
					      print_cur,
					      print_next,
					      cond);	/* Modified! */
	    temp1 = bdd_and(gbm,example,sub_example);
	    bdd_free(gbm,example);
	    bdd_free(gbm,sub_example);
	    example = temp1;
	    offset += temp->field_type->bdd_count;
	    temp = temp->next;
	}
	return example;
    }
}


bdd bdd_print_example(bdd cond)
{
    string_list name_list;
    bdd example, sub_example, remaining_cond, temp1;
    variable_desig temp_desig;
    int prev_phase, cur_phase, next_phase;

    if (cond==bdd_zero(gbm)) {
	printf("*** Error:  No Possible Example State ***\n");
	return bdd_zero(gbm);
    }

    /* Decide which phases to print. */
    bdd_assoc(gbm,prev_vars_assoc);
    temp1 = bdd_exists(gbm,cond);
    prev_phase = (temp1!=cond);
    bdd_free(gbm,temp1);

    bdd_assoc(gbm,cur_vars_assoc);
    temp1 = bdd_exists(gbm,cond);
    cur_phase = (temp1!=cond);
    bdd_free(gbm,temp1);

    bdd_assoc(gbm,next_vars_assoc);
    temp1 = bdd_exists(gbm,cond);
    next_phase = (temp1!=cond);
    bdd_free(gbm,temp1);

    /* Step through all the variables, printing each one. */
    name_list = global_variable_names;
    example = bdd_one(gbm);
    remaining_cond = bdd_identity(gbm,cond);
    while (name_list!=NULL) {
	printf("%s",name_list->data);
	temp_desig = get_variable_descriptor(name_list->data);
	/* FIX THIS need to print dependent variables.  AJH Oct 16, 1992 */
	if (temp_desig->fdv) {
		printf(" is dependent (FIX THIS)\n");
	} else {	/* temporary hack to skip this section for fdv */
	sub_example = bdd_partial_example(temp_desig->bits,	/* base */
					  1,	/* depth of indentation */
					  temp_desig->base_type,
					  0,	/* offset */
					  prev_phase,
					  cur_phase,
					  next_phase,
					  &remaining_cond);	/* Modified! */
	temp1 = bdd_and(gbm,example,sub_example);
	    bdd_free(gbm,example);
	    bdd_free(gbm,sub_example);
	example = temp1;
	} /* end of temporary hack to skip this section for fdv */

	free(temp_desig);
	name_list = name_list->next;
    }

    bdd_free(gbm,remaining_cond);

    return example;
}



bdd select_small_frontier(bdd c0, bdd c1)
	/* Tries to find a set X, such that c1-c0 <= X <= c1, and
	   the BDD for X is small. */
{
	bdd b0;
	bdd temp1;
	bdd gcf;
	long whole_size, b0_size, gcf_size;

	temp1 = bdd_not(gbm,c0);
	b0 = bdd_and(gbm,c1,temp1);
	gcf = simplify_bdd(c1,temp1);
	/*gcf = bdd_identity(gbm,b0);*/	/* Temporary speed hack -- DEBUG! AJH*/
	if (b0==gcf) printf("b0=gcf\n");
	else printf("b0 != gcf\n");
	bdd_free(gbm,temp1);

	printf("Sizes:  ");
	printf("\twhole = %ld",whole_size = bdd_size(gbm,c1,1));
	printf("\tdiff = %ld",b0_size = bdd_size(gbm,b0,1));
	printf("\tgcf = %ld\n",gcf_size = bdd_size(gbm,gcf,1));

	/* Choose the smallest frontier set. */
	if (whole_size > b0_size) {
		if (b0_size > gcf_size) {
			/* gcf is smallest */
			bdd_free(gbm,b0);
			return gcf;
		} else {
			/* diff is smallest */
			bdd_free(gbm,gcf);
			return b0;
		}
	} else {
		if (whole_size > gcf_size) {
			/* gcf is smallest */
			bdd_free(gbm,b0);
			return gcf;
		} else {
			/* whole is smallest */
			bdd_free(gbm,b0);
			bdd_free(gbm,gcf);
			return bdd_identity(gbm,c1);
		}
	}
}



bdd bdd_print_trace(bdd previous,
			    bdd start,
			    uneval_prop next,
			    bdd goal)
{
    bdd example, next_example;
    bdd temp1, temp2, temp3;
    bdd new;

    if (global_fdv_count>0) {
	/* Must check that reachable states won't allow violation
	   of funtional dependencies. */
	temp1 = check_MOB(start);	/* Try to find possible violation. */
	if (temp1!=bdd_zero(gbm)) {
	    /* Functionally dependency violated for some variable. */
	    printf("A REACHABLE STATE LEADING TO DEPENDENCY VIOLATION:\n");
	    example = bdd_print_example(temp1);
	    bdd_free(gbm,temp1);
	    return(example);
	}
	bdd_free(gbm,temp1);
    }

    temp1 = bdd_and(gbm,start,goal);	/* Have we reached goal? */
    if (temp1!=bdd_zero(gbm)) {
	printf("A POSSIBLE END STATE (invariant violated/goal reached):\n");
	example = bdd_print_example(temp1);
	    bdd_free(gbm,temp1);
    } else {
	    bdd_free(gbm,temp1);
	/* URRRRP!!!  This is potentially non-monotonic!  --AJH 9/10/95 */
	/* Double-Urp!  This is actually OK.  "start" is actually
		the preceding set, not the initial start set. --AJH 9/24/95 */
	temp2 = select_small_frontier(previous,start);
	/*temp2 = bdd_identity(gbm,start);*/	/* Temp hack 9/10/95 */
	temp1 = compute_forward_image(temp2,next,FALSE);
	    bdd_free(gbm,temp2);
	/* Convert from next to current phase. */
	bdd_assoc(gbm,next_vars_assoc);
	temp2 = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
	    bdd_free(gbm,temp1);

	new = bdd_or(gbm,start,temp2);
	    bdd_free(gbm,temp2);
	/*bdd_print_profile(gbm,new,bdd_naming_fn_full,NULL,80,stdout);*/
	printf("(%ld)",bdd_size(gbm,new,1)); fflush(stdout);
	if (start==new) {
	    /*printf("Completed reachability:  %lf states found.\n",
			enumerate_reachable_states(new,global_bdd_count/3));*/
			/* This can be very time-consuming. */
	    printf("A POSSIBLE END STATE (farthest reachable):\n");
	    temp1 = bdd_not(gbm,previous);
	    temp2 = bdd_and(gbm,start,temp1);
		bdd_free(gbm,temp1);
	    example = bdd_print_example(temp2);
		bdd_free(gbm,temp2);
	} else {
	    next_example = bdd_print_trace(start,new,next,goal);

	    /* Convert next_example to next phase. */
	    bdd_assoc(gbm,cur_vars_assoc);
	    temp2 = bdd_rel_prod(gbm,next_example,current_next_eq->bdd);
		bdd_free(gbm,next_example);
	    temp1 = compute_backward_image(temp2,next);
		bdd_free(gbm,temp2);
	    temp2 = bdd_and(gbm,start,temp1); /* Ensure example is reachable. */
		bdd_free(gbm,temp1);
	    printf("A POSSIBLE PREDECESSOR STATE:\n");
	    example = bdd_print_example(temp2);
	    bdd_free(gbm,temp2);
	}
	bdd_free(gbm,new);
    }

    return example;
}


void printtrace(uneval_prop startstates,
		uneval_prop nextstate,
		uneval_prop goal)
{
    printf("Start of printtrace.\n");
    reset_memory_meter();
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating start condition...");fflush(stderr);
    evaluate_proposition(startstates,FALSE);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating start condition.\n");
    printf("Start condition size = %ld or %ld\n",
			bdd_size(gbm,startstates->bdd,1),
			count_bdd_nodes(startstates));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating goal/invariant...");fflush(stderr);
    evaluate_proposition(goal,FALSE);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating goal/invariant.\n");
    printf("Goal/invariant size = %ld or %ld\n",
			bdd_size(gbm,goal->bdd,1),
			count_bdd_nodes(goal));
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    if (global_fdv_count>0) {
	/* If there are any fdv's, we must build multiset of BFVs. */
	fprintf(stderr,"Building MOB for nextstate relation...");fflush(stderr);
	build_MOB(nextstate);	/* Creates global, static MOB. */
	fprintf(stderr,"\n");fflush(stderr);
    }

    fprintf(stderr,"Starting reachability...");fflush(stderr);
    bdd_free(gbm,
	bdd_print_trace(bdd_zero(gbm),
			startstates->bdd,
			nextstate,
			goal->bdd)
    );	/* Get rid of extra example. */
    fprintf(stderr,"\n");fflush(stderr);

    printf("End of printtrace.\n");
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Next state relation uses %ld bdd nodes.\n",
					count_bdd_nodes(nextstate));
    printf("\n");
    fflush(stdout);
}

void quick_and_dirty_disjunct_list_simplify(disjunct_list x)
{
	int i,j;
	bdd temp1, temp2;
	long isize, jsize;

	for (j=0; j < x->size; j++) {
	    temp1 = bdd_not(gbm,x->disjunct[j]);
	    jsize = bdd_size(gbm,x->disjunct[j],1);
	    for (i=0; i < x->size; i++) {
		if (i!=j) {
		    isize = bdd_size(gbm,x->disjunct[i],1);
		    if (isize >= jsize) {
			/*fprintf(stderr,"Simplify %d(%ld) by %d(%ld) --",
			    i,isize,j,jsize); fflush(stderr);*/
			temp2 = simplify_bdd(x->disjunct[i],temp1);
			bdd_free(gbm,x->disjunct[i]);
			x->disjunct[i] = temp2;
			/*fprintf(stderr," Done.\n");*/
		    } else {
			/*fprintf(stderr," Skipped.\n");*/
		    }
		}
	    }
	    bdd_free(gbm,temp1);
	}
}

void print_fwd_counterexample_trace(conjunct_list good,
					uneval_prop next,
					int depth)
{
    int which_disj, which_conj;
    int i;
    bdd bad_set, pre_image, example;

    /* Find which disjunct violates which conjunct. */
    find_violation_id_ic(R[depth],good,&which_disj,&which_conj);
    if (which_disj == -1) {
	fprintf(stderr,"ERROR!!!!!  print_fwd_counterexample_trace() called with no possible trace!\n");
	return;
    }

    printf("THE FOLLOWING STATE VIOLATES PROPERTY #%d:\n",which_conj);

    /* Get a set that contains a counterexample.  This is guaranteed
	non-empty since find_violation_id_ic came back successfully. */
    bad_set = bdd_implies(gbm,R[depth]->disjunct[which_disj],
				good->conjunct[which_conj]);
    for (i=depth; i>0; i--) {
	/* Loop Invariant:  On each iteration at this point,
	   bad_set contains a subset of R[i] that leads to a violation. */

	/* Choose an example state and print it. */
	example = bdd_print_example(bad_set);
	bdd_free(gbm,bad_set);

	printf("WHICH HAS PREDECESSOR:\n");

	pre_image = pre_image_bdd(next,example);
	bdd_free(gbm,example);

	/* Find a disjunct in R[i-1] that could have led to counterexample. */
	for (which_disj=0; which_disj < R[i-1]->size; which_disj++) {
	    bad_set = bdd_intersects(gbm,pre_image,
				R[i-1]->disjunct[which_disj]);
		/* David Longism:  bdd_intersects finds a small bdd that
			implies the intersection, but is FALSE iff the
			intersection is empty. */
	    if (bad_set!=bdd_zero(gbm)) break;	/* We've found it. */
	    bdd_free(gbm,bad_set);
	}
	bdd_free(gbm,pre_image);
    }

    /* Print the start state. */
    example = bdd_print_example(bad_set);
    bdd_free(gbm,bad_set);
    bdd_free(gbm,example);
    printf("WHICH IS A START STATE.\n");
}

void bdd_check_invariant_fwd(disjunct_list start,
			uneval_prop next,
			conjunct_list good)
{
    int depth = 0;
    int i, j;
    disjunct_list new_list;

    R[0] = copy_disjunct_list(start);

    do {
	print_disjunct_list_size(R[depth],depth); /* For benchmarking only. */
	if (!check_disjunct_list_implies_conjunct_list(R[depth],good)) {
	    /* We have a violation.  Print counterexample trace. */
	    print_fwd_counterexample_trace(good,next,depth);

	    /* Free intermediate reachable state sets. */
	    for (i=depth; i>=0; i--) {
		free_disjunct_list(R[i]);
	    }
	    return;
	}

	new_list = image_id(next,R[depth]);
	/* FIX THIS -- NEED TO SIMPLIFY */
	R[depth+1] = concat_disjunct_lists(R[0],new_list);
	free_disjunct_list(new_list);
	fast_munch_disjunct_list(R[depth+1]);
	medium_munch_disjunct_list(R[depth+1]);
	quick_and_dirty_disjunct_list_simplify(R[depth+1]);
	fast_munch_disjunct_list(R[depth+1]);
	medium_munch_disjunct_list(R[depth+1]);
	/*phased_slow_munch_disjunct_list(R[depth+1],0.8,2);*/
	minmax_munch_disjunct_list(R[depth+1]);

	depth++;
    } while (!check_disjunct_list_implication(R[depth],R[depth-1]));
    print_disjunct_list_size(R[depth],depth); /* For benchmarking only. */

    /* Everything's OK. */
    printf("INVARIANT IS CORRECT.\n");

    /* Free intermediate reachable state sets. */
    for (i=depth; i>=0; i--) {
	free_disjunct_list(R[i]);
    }
} /* bdd_check_invariant_fwd */

/* Goes forward, but uses implicit disjunctions. */
void check_invariant_fwd(uneval_prop startstates,
			 uneval_prop nextstate,
			 uneval_prop properties)
{
    conjunct_list eval_properties;
    disjunct_list eval_start;

    printf("Start of check_invariant_fwd.\n");
    reset_memory_meter();
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating start condition...");fflush(stderr);
    eval_start = eval_startstates(startstates);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating start condition.\n");
    printf("Start state BDD list sizes:\n");
    print_disjunct_list_size(eval_start,0);
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    fprintf(stderr,"Evaluating properties...");fflush(stderr);
    eval_properties = eval_invariant(properties);
    fprintf(stderr,"\n");fflush(stderr);

    printf("Done evaluating invariant.\n");
    printf("Properties BDD list sizes:\n");
    print_conjunct_list_size(eval_properties,0);
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("\n");
    fflush(stdout);

    if (global_fdv_count>0) {
	/* I don't support FDVs with check_invariant.  Bail out. */
	fprintf(stderr,"Error:  Cannot use dependent variables with checkinvariantfwd.\n");
	return;
    }

    fprintf(stderr,"Starting check...");fflush(stderr);
    bdd_check_invariant_fwd(eval_start,
			nextstate,
			eval_properties);
    free_disjunct_list(eval_start);
    free_conjunct_list(eval_properties);
    fprintf(stderr,"\n");fflush(stderr);

    printf("End of check_invariant_fwd.\n");
    printf("Clock %d/100 seconds.\n",user_cpu_time()*5/3);
    printf("Memory %ld bytes.\n",read_memory_meter());
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Garbage Collecting...\n");
    bdd_gc(gbm);
    printf("BDD %ld nodes.\n",bdd_total_size(gbm));
    printf("Next state relation uses %ld bdd nodes.\n",
					count_bdd_nodes(nextstate));
    printf("\n");
    fflush(stdout);
} /* check_invariant_fwd */
