/* Copyright 1993 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Routines For Implicit Conjunctions and Disjunctions. */
/* By Alan J. Hu */
/* January 4, 1993 */
/* Copied into separate file on September 11, 1994. */

/* ************************ CHANGE LOG ************************** */
/*
7 September 1994 -- Moved CTL stuff to file everctl.c.
		    Moved basic conjunct and disjunct list stuff to everlist.c.
		    Moved image computation stuff to everimage.c.
*/
/* ************************************************************** */

#include "ever.h"
#include "everlist.h"	/* Stuff for conjunct and disjunct lists. */
#include "bddajh.h"

bdd simplify_bdd(bdd f, bdd c);	/* In everinv.c. */



/* Functions to work on conjunct_list */

conjunct_list new_conjunct_list(int size)
{
	conjunct_list new;

	new = (conjunct_list)malloc(sizeof(struct conjunct_list_node));
	new->size = size;
	new->conjunct = (bdd *)calloc(size,sizeof(bdd));

	return new;
}


disjunct_list new_disjunct_list(int size)
{
	return (disjunct_list)new_conjunct_list(size);
}

void free_conjunct_list(conjunct_list x)
{
	int i;

	for (i = 0; i < x->size; i++) {
	    if (x->conjunct[i]!=NULL) {
		bdd_free(gbm,x->conjunct[i]);
	    }
	}
	free(x->conjunct);
	free(x);
}

void free_disjunct_list(disjunct_list x)
{
	free_conjunct_list((conjunct_list)x);
}

conjunct_list copy_conjunct_list(conjunct_list x)
{
	conjunct_list y;
	int i;

	y = new_conjunct_list(x->size);
	for (i=0; i < x->size; i++) {
		y->conjunct[i] = bdd_identity(gbm,x->conjunct[i]);
	}

	return y;
}

disjunct_list copy_disjunct_list(disjunct_list x)
{
	return (disjunct_list)copy_conjunct_list((conjunct_list)x);
}

conjunct_list concat_conjunct_lists(conjunct_list a, conjunct_list b)
{
	conjunct_list out;
	int i;

	out = new_conjunct_list(a->size + b->size);
	for (i=0; i < a->size; i++)
		out->conjunct[i] = bdd_identity(gbm,a->conjunct[i]);
	for (i=0; i < b->size; i++)
		out->conjunct[a->size + i] = bdd_identity(gbm,b->conjunct[i]);

	return out;
}

disjunct_list concat_disjunct_lists(disjunct_list a, disjunct_list b)
{
	return (disjunct_list)concat_conjunct_lists(
			(conjunct_list)a, (conjunct_list)b);
}

conjunct_list or_conjunct_lists(conjunct_list a, conjunct_list b)
{
	conjunct_list c;
	int i,j;

	c = new_conjunct_list(a->size * b->size);
	for (i=0; i < a->size; i++)
	    for(j=0; j < b->size; j++)
		c->conjunct[i*(b->size)+j] = bdd_or(gbm,a->conjunct[i],
							b->conjunct[j]);

	return c;
}

void append_to_conjunct_list(bdd x, conjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
{
	a->size++;
	a->conjunct = (bdd *)realloc(a->conjunct,
					a->size*sizeof(bdd));
	a->conjunct[a->size-1] = bdd_identity(gbm,x);
}

conjunct_list de_morgan_dl_to_cl(disjunct_list x)
	/* Returns (not x) using de Morgan's Law. */
{
	int i;
	conjunct_list result;

	result = new_conjunct_list(x->size);
	for (i=0; i < x->size; i++) {
		result->conjunct[i] = bdd_not(gbm,x->disjunct[i]);
	}

	return result;
}

disjunct_list de_morgan_cl_to_dl(conjunct_list x)
	/* Returns (not x) using de Morgan's Law. */
{
	return (disjunct_list)de_morgan_dl_to_cl((disjunct_list)x);
}

void de_morgan_in_place(conjunct_list x)
	/* Complements the components of x. */
{
	int i;
	bdd temp;

	for (i=0; i < x->size; i++) {
		temp = bdd_not(gbm,x->conjunct[i]);
		bdd_free(gbm,x->conjunct[i]);
		x->conjunct[i] = temp;
	}
}

void fast_munch_conjunct_list(conjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* A very quick and dirty attempt to shorten the conjunct_list
	by looking for TRUEs and FALSEs. */
{
	int old,new,i;

	/* Scan list for a FALSE conjunct. */
	for (old=0; old < a->size; old++) {
	    if (a->conjunct[old]==bdd_zero(gbm)) {
		/* There's a FALSE in the list.  The whole thing is FALSE. */
		for (i=0; i < a->size; i++) {
		    bdd_free(gbm, a->conjunct[i]);
		}
		/* Change the list to have only one element. */
		a->conjunct = (bdd *)realloc(a->conjunct,sizeof(bdd));
		a->size = 1;
		a->conjunct[0] = bdd_zero(gbm);

		return;		/* BAIL OUT. */
	    }
	}

	/* Skip TRUEs from list. */
	for (old=new=0; old < a->size; old++) {
	    if (a->conjunct[old]==bdd_one(gbm)) {
		/* It's a TRUE.  Skip it. */
		bdd_free(gbm,a->conjunct[old]);	/* Not necessary. */
	    } else {
		/* Must copy it. */
		a->conjunct[new] = a->conjunct[old];
		new++;
	    }
	}

	/* Patch up if list ends up empty. */
	if (new==0) {
	    new=1;
	    a->conjunct[0] = bdd_one(gbm);
	}

	a->conjunct = (bdd *)realloc(a->conjunct,sizeof(bdd)*new);
	a->size = new;
}

void medium_munch_conjunct_list(conjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Looks for duplicate conjuncts or complementary conjuncts. */
{
	int ok;
	int i,j;
	bdd negation;

	ok = 0;
	/* Loop invariant:  a->conjunct[0..ok-1] have no duplicates or
		complements, a->conjunct[ok..i-1] have been freed, and
		a->conjunct[i..end] are as in original. */
	for (i=0; i < a->size; i++) {
	    /* Compare conjunct[i] against conjunct[0..ok-1] */
	    negation = bdd_not(gbm,a->conjunct[i]);
	    for (j=0; j < ok; j++) {

		/* Check complement. */
			/* I can optimize this by using BDD_SAME_OR_NEGATION
				from bddint.h in David Long's packages. */
		if (a->conjunct[j] == negation) {
		    /* Conjunct list has complement.  BAIL OUT!!! */
		    int k;

		    /* Free everything. */
		    bdd_free(gbm,negation);
		    for (k=0; k < ok; k++) bdd_free(gbm,a->conjunct[k]);
		    for (k=i; k < a->size; k++) bdd_free(gbm,a->conjunct[k]);

		    /* Redo the conjunct list. */
		    a->conjunct = (bdd *)realloc(a->conjunct,sizeof(bdd)*1);
		    a->conjunct[0] = bdd_zero(gbm);
		    a->size = 0;

		    return;	/* EXIT! */
		}

		/* Check duplicate */
		if (a->conjunct[j] == a->conjunct[i]) {
		    bdd_free(gbm,a->conjunct[i]);
		    a->conjunct[i] = NULL;
		    break;	/* exit loop on j */
		}

	    } /* for j */
	    bdd_free(gbm,negation);

	    if (a->conjunct[i]!=NULL) {
		a->conjunct[ok] = a->conjunct[i];
		ok++;
	    }
	} /* for i */

	a->conjunct = (bdd *)realloc(a->conjunct,sizeof(bdd)*ok);
	a->size = ok;
}

void phased_slow_munch_conjunct_list(conjunct_list a,
					float grow_threshold,
					int number_of_phases)
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Tries to shorten conjunct list a. */
    /* This is slightly greedier than slow_munch_conjunct_list.
       It goes through several phases.  At phase i, it uses a
       grow_threshold of grow_threshold*i/number_of_phases.
       During a phase, it combines any conjuncts that gives a growth
       ratio less than the current grow_threshold (rather than looking
       for the minimum growth ratio). */
{
    int phase;
    int i,j;
    int whichi, whichj;
    bdd conjunction;
    float ratio;
    bdd size_temp[3];
    int size;
    int actual_size;
    float *pair_table;
    float UNKNOWN = -1.0;	/* Indicates an unknown entry in the table. */
    long additive_size;
    float phased_threshold;
    int found_flag;

    /* Initialize conjunction ratio table. */
    size = a->size;
    actual_size = size;
    pair_table = (float *)calloc(size*size,sizeof(float));
    size_temp[2] = NULL;
    for (i=0; i < size; i++) {
	for (j=i+1; j < size; j++) {
	    pair_table[i*size+j] = UNKNOWN;
	}
    }

    for (phase = 1; phase <= number_of_phases; phase++) {
	phased_threshold = grow_threshold * phase / number_of_phases;
	fprintf(stderr,"Phase %d, Threshold = %f.\n",phase,phased_threshold);

	do {
	    found_flag = FALSE;
	    /* Look for a conjunction that gives a growth ratio smaller
		than phased_threshold. */
	    for (i=0; (i<actual_size) && (!found_flag); i++) {
		for (j=i+1; (j<actual_size) && (!found_flag); j++) {
		    if (pair_table[i*size+j]!=UNKNOWN) {
			/* We've evaluated this pair before. */
			if (pair_table[i*size+j]<phased_threshold) {
			    /* It's good.  Rebuild the conjunction. */
			    found_flag = TRUE;
			    whichi = i;
			    whichj = j;
			    ratio = pair_table[i*size+j];
			    conjunction = bdd_and(gbm,a->conjunct[i],
							a->conjunct[j]);
			} else {
			    /* Ratio is bad.  Ignore it. */
			}
		    } else {
			/* We've never evaluated this pair before. */
			size_temp[0] = a->conjunct[i];
			size_temp[1] = a->conjunct[j];
			additive_size = bdd_size_multiple(gbm,size_temp,1);
			/*additive_size = bdd_size(gbm,size_temp[0],1) +
					    bdd_size(gbm,size_temp[1],1);*/

			conjunction = bdd_bounded_ite(gbm,a->conjunct[i],
							a->conjunct[j],
							bdd_zero(gbm),
					(long)(additive_size*grow_threshold+1));

			if (conjunction == NULL) {
			    /* conjunction overflowed bound */
			    ratio = grow_threshold+2;
			} else {
			    ratio = (float)bdd_size(gbm,conjunction,1)
							/additive_size;
			    if (ratio<phased_threshold) {
				found_flag = TRUE;
				whichi = i;
				whichj = j;
			    } else {
				bdd_free(gbm,conjunction);
			    }
			}
			pair_table[i*size+j] = ratio;
		    }
		}
	    }
	    /* If found_flag is true, whichi, whichj, ratio, and conjunction
		should all be valid. */

	    /* If we found a conjunction to evaluate, do it. */
	    if (found_flag) {
		fprintf(stderr,"Combining %d and %d ratio = %f, array_size = %d.\n",whichi,whichj,ratio,actual_size);
		bdd_free(gbm,a->conjunct[whichi]);
		bdd_free(gbm,a->conjunct[whichj]);
		a->conjunct[whichi] = conjunction;
		a->conjunct[whichj] = a->conjunct[--actual_size];
		/* Update the conjunction ratio table. */
		/* This is kind of tricky.  We've moved the last conjunct
			into the place of whichj, so we must copy the last
			column of the table into the whichj position, and
			we must also flag anything related to the whichi
			position as being unknown. */
		for (i=0; i<whichj; i++)
		    pair_table[i*size+whichj] = pair_table[i*size+actual_size];
		for (j=whichj+1; j<actual_size; j++)
		    pair_table[whichj*size+j] = pair_table[j*size+actual_size];
		for (i=0; i<whichi; i++) pair_table[i*size+whichi] = UNKNOWN;
		for (j=whichi+1; j<actual_size; j++)
					pair_table[whichi*size+j] = UNKNOWN;
	    }
	} while (found_flag);
    }

    /* Resize a. */
    a->size = actual_size;
    a->conjunct = (bdd *)realloc(a->conjunct,actual_size*sizeof(bdd));
    
    /* Free the pair table. */
    free(pair_table);

    return;
} /* phased_slow_munch_conjunct_list */

void slow_munch_conjunct_list(conjunct_list a, float grow_threshold)
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Tries to shorten conjunct list a. */
    /* Implements the greedy heuristic from DAC 94 that always combines
	the pair which results in the least growth. */
    /* This implementation should use less space, and probably run faster
	than what I had used in evaluate_some_conjunctions(). */
{
    int i,j;
    int besti, bestj;
    float ratio,bestratio;
    bdd size_temp[3];
    bdd conjunction;
    int size;		/* initial size of a (and corresponding tables) */
    int actual_size;	/* current size of list after shrinkage */
    float *pair_table;
    long additive_size;

    /* Initialize pairwise conjunction size table. */
    size = a->size;
    fprintf(stderr,"Building %d x %d conjunction size table...\n",size,size);
    actual_size = size;
    pair_table = (float *)calloc(size*size,sizeof(float));
    size_temp[2] = NULL;
    for (i=0; i<size; i++) {
	fprintf(stderr,"Ghee-");fflush(stderr);
	for (j=i+1; j<size; j++) {
	    size_temp[0] = a->conjunct[i];
	    size_temp[1] = a->conjunct[j];

	    additive_size = bdd_size_multiple(gbm,size_temp,1);
	    /*additive_size = bdd_size(gbm,size_temp[0],1) +
				bdd_size(gbm,size_temp[1],1);*/

	    /*fprintf(stderr,"Ghee-");fflush(stderr);*/
	    conjunction = bdd_bounded_ite(gbm,a->conjunct[i],a->conjunct[j],bdd_zero(gbm),(long)(additive_size*grow_threshold+1));
	    /*conjunction = bdd_and(gbm,a->conjunct[i],a->conjunct[j]);*/
	    /*fprintf(stderr,"bo.");fflush(stderr);*/

	    if (conjunction == NULL) {
		/* conjunction overflowed bound */
		/*fprintf(stderr,"@"); fflush(stderr);*/
		ratio = grow_threshold+2;
	    } else {
		ratio = (float)bdd_size(gbm,conjunction,1)/additive_size;
		bdd_free(gbm,conjunction);
		    /* Maybe save it if ratio<grow_threshold? */
	    }
	    pair_table[i*size+j] = ratio;
	}
	fprintf(stderr,"bo.");fflush(stderr);
    }

    /* Combine Terms Until No More Improvement. */
    while (1) {
	/* Find best pair to combine. */
	bestratio = grow_threshold+1;
	for (i=0; i<actual_size; i++)
	    for (j=i+1; j<actual_size; j++) {
		ratio = pair_table[i*size+j];
		if (ratio<bestratio) {
		    bestratio = ratio;
		    besti = i;
		    bestj = j;
		}
	    }
	fprintf(stderr,"Best ratio is %lg combining %d and %d.  Array size = %d.\n",bestratio,besti,bestj,actual_size);

	if (bestratio > grow_threshold) break;	/* Loop Exit */

	/* Update the pairwise conjunction table. */
	/* Do the conjunction. */
	conjunction = bdd_and(gbm,a->conjunct[besti],a->conjunct[bestj]);
	bdd_free(gbm,a->conjunct[besti]);
	bdd_free(gbm,a->conjunct[bestj]);
	a->conjunct[besti] = conjunction;
	/* Delete anything related to bestj. */
	for (i=bestj+1; i<actual_size; i++)
	    a->conjunct[i-1] = a->conjunct[i];
	for (i=0; i<actual_size; i++)
	    for (j=bestj+1; j<actual_size; j++)
		if (i<=bestj) {
		    /* Only need to slide early rows to the left. */
		    pair_table[i*size+j-1] = pair_table[i*size+j];
		} else {
		    /* Rows below bestj need to be slid up as well. */
		    pair_table[(i-1)*size+j-1] = pair_table[i*size+j];
		}
	actual_size--;
	/* Compute the new entries related to besti */
	for (i=0; i<besti; i++) {
	    size_temp[0] = a->conjunct[i];
	    size_temp[1] = a->conjunct[besti];

	    additive_size = bdd_size_multiple(gbm,size_temp,1);
	    /*additive_size = bdd_size(gbm,size_temp[0],1) +
				bdd_size(gbm,size_temp[1],1);*/

	    /*conjunction = bdd_and(gbm,a->conjunct[i],a->conjunct[besti]);*/
	    conjunction = bdd_bounded_ite(gbm,a->conjunct[i],a->conjunct[besti],bdd_zero(gbm),(long)(additive_size*grow_threshold+1));

	    if (conjunction == NULL) {
		/* conjunction overflowed bound */
		ratio = grow_threshold+2;
	    } else {
		ratio = (float)bdd_size(gbm,conjunction,1)/additive_size;
		bdd_free(gbm,conjunction);
		    /* Maybe save it if ratio<grow_threshold? */
	    }
	    pair_table[i*size+besti] = ratio;
	}
	for (j=besti+1; j<actual_size; j++) {
	    size_temp[0] = a->conjunct[besti];
	    size_temp[1] = a->conjunct[j];

	    additive_size = bdd_size_multiple(gbm,size_temp,1);
	    /*additive_size = bdd_size(gbm,size_temp[0],1) +
				bdd_size(gbm,size_temp[1],1);*/

	    /*conjunction = bdd_and(gbm,a->conjunct[besti],a->conjunct[j]);*/
	    conjunction = bdd_bounded_ite(gbm,a->conjunct[besti],a->conjunct[j],bdd_zero(gbm),(long)(additive_size*grow_threshold+1));

	    if (conjunction == NULL) {
		/* conjunction overflowed bound */
		ratio = grow_threshold+2;
	    } else {
		ratio = (float)bdd_size(gbm,conjunction,1)/additive_size;
		bdd_free(gbm,conjunction);
		    /* Maybe save it if ratio<grow_threshold? */
	    }
	    pair_table[besti*size+j] = ratio;
	}
    }

    /* Resize a. */
    a->size = actual_size;
    a->conjunct = (bdd *)realloc(a->conjunct,actual_size*sizeof(bdd));
    
    /* Free the pair table. */
    free(pair_table);

    return;
} /* slow_munch_conjunct_list */

void minmax_munch_conjunct_list(conjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
	/* Combines conjuncts if the size of the result is no larger than
	    the size of the larger of the two. */
{
    int i,j;
    int whichi, whichj;
    bdd conjunction;
    bdd size_temp[3];
    int size;
    int actual_size;
    long *pair_table;
    long UNKNOWN = -1;	/* Indicates an unknown entry in the table. */
    long larger_size;
    long result_size;
    int found_flag;

    /* Initialize conjunction size table. */
    size = a->size;
    actual_size = size;
    pair_table = (long *)calloc(size*size,sizeof(long));
    size_temp[2] = NULL;
    for (i=0; i < size; i++) {
	for (j=i+1; j < size; j++) {
	    pair_table[i*size+j] = UNKNOWN;
	}
    }
    for (i=0; i < size; i++) {
	pair_table[i*size+i] = bdd_size(gbm,a->conjunct[i],1);
    }

    /* Repeat as long as we've combined something. */
    do {
	found_flag = FALSE;
	/* Look for a conjunction that gives a result smaller
	    than the larger of the two arguments. */
	for (i=0; (i<actual_size) && (!found_flag); i++) {
	    for (j=i+1; (j<actual_size) && (!found_flag); j++) {
		/* Compute the larger of the two BDDs. */
		if (pair_table[i*size+i] > pair_table[j*size+j]) {
		    larger_size = pair_table[i*size+i];
		} else {
		    larger_size = pair_table[j*size+j];
		}

		if (pair_table[i*size+j]!=UNKNOWN) {
		    /* We've evaluated this pair before. */
		    if (pair_table[i*size+j] <= 1.5*larger_size) {
			/* It's good.  Rebuild the conjunction. */
			found_flag = TRUE;
			whichi = i;
			whichj = j;
			result_size = pair_table[i*size+j];
			conjunction = bdd_and(gbm,a->conjunct[i],
						    a->conjunct[j]);
		    } else {
			/* Ratio is bad.  Ignore it. */
		    }
		} else {
		    /* We've never evaluated this pair before. */
		    conjunction = bdd_bounded_ite(gbm,a->conjunct[i],
						    a->conjunct[j],
						    bdd_zero(gbm),
						    larger_size+1);
		    if (conjunction == NULL) {
			/* conjunction overflowed bound */
			result_size = larger_size+1000;	/* Ensure too big. */
		    } else {
			result_size = bdd_size(gbm,conjunction,1);
			if (result_size <= 1.5*larger_size) {
			    found_flag = TRUE;
			    whichi = i;
			    whichj = j;
			} else {
			    bdd_free(gbm,conjunction);
			}
		    }
		    pair_table[i*size+j] = result_size;
		}
	    }
	}
	/* If found_flag is true, whichi, whichj, result_size, and conjunction
	    should all be valid. */

	/* If we found a conjunction to evaluate, do it. */
	if (found_flag) {
	    fprintf(stderr,"Combining %d and %d result = %ld, array_size = %d.\n",whichi,whichj,result_size,actual_size);
	    bdd_free(gbm,a->conjunct[whichi]);
	    bdd_free(gbm,a->conjunct[whichj]);
	    a->conjunct[whichi] = conjunction;
	    a->conjunct[whichj] = a->conjunct[--actual_size];
	    /* Update the conjunction size table. */
	    /* This is kind of tricky.  We've moved the last conjunct
		    into the place of whichj, so we must copy the last
		    column of the table into the whichj position, and
		    we must also flag anything related to the whichi
		    position as being unknown. */
	    for (i=0; i<whichj; i++)
		pair_table[i*size+whichj] = pair_table[i*size+actual_size];
	    for (j=whichj+1; j<actual_size; j++)
		pair_table[whichj*size+j] = pair_table[j*size+actual_size];
	    for (i=0; i<whichi; i++) pair_table[i*size+whichi] = UNKNOWN;
	    for (j=whichi+1; j<actual_size; j++)
				    pair_table[whichi*size+j] = UNKNOWN;
	}
    } while (found_flag);

    /* Resize a. */
    a->size = actual_size;
    a->conjunct = (bdd *)realloc(a->conjunct,actual_size*sizeof(bdd));
    
    /* Free the pair table. */
    free(pair_table);

    return;
}

void fast_munch_disjunct_list(disjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
{
	de_morgan_in_place((conjunct_list)a); /* Negation is conjunct_list. */
	fast_munch_conjunct_list((conjunct_list)a);
	de_morgan_in_place((conjunct_list)a); /* Make it a disjunct_list. */
}

void medium_munch_disjunct_list(disjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
{
	de_morgan_in_place((conjunct_list)a); /* Negation is conjunct_list. */
	medium_munch_conjunct_list((conjunct_list)a);
	de_morgan_in_place((conjunct_list)a); /* Make it a disjunct_list. */
}

void phased_slow_munch_disjunct_list(disjunct_list a,
					float grow_threshold,
					int number_of_phases)
	/* DESTRUCTIVE LIST MANIPULATION! */
{
	de_morgan_in_place((conjunct_list)a); /* Negation is conjunct_list. */
	phased_slow_munch_conjunct_list((conjunct_list)a,
					grow_threshold,
					number_of_phases);
	de_morgan_in_place((conjunct_list)a); /* Make it a disjunct_list. */
}

void slow_munch_disjunct_list(disjunct_list a, float grow_threshold)
	/* DESTRUCTIVE LIST MANIPULATION! */
{
	de_morgan_in_place((conjunct_list)a); /* Negation is conjunct_list. */
	slow_munch_conjunct_list((conjunct_list)a, grow_threshold);
	de_morgan_in_place((conjunct_list)a); /* Make it a disjunct_list. */
}

void minmax_munch_disjunct_list(disjunct_list a)
	/* DESTRUCTIVE LIST MANIPULATION! */
	/* Combines disjuncts if the size of the result is no larger than
	    the size of the larger of the two. */
{
	de_morgan_in_place((conjunct_list)a); /* Negation is conjunct_list. */
	minmax_munch_conjunct_list((conjunct_list)a);
	de_morgan_in_place((conjunct_list)a); /* Make it a disjunct_list. */
}




bdd simplify_bdd_by_conjunct_list(bdd f, conjunct_list c)
{
	int i;
	bdd temp1, temp2;

	printf("\nSimplifying!\n");
	for (i=0; i < c->size; i++)printf("%ld ",bdd_size(gbm,c->conjunct[i],1));
	printf("\n");

	temp1 = bdd_identity(gbm,f);
	printf("%ld",bdd_size(gbm,temp1,1));
	for (i=0; i < c->size; i++) {
		temp2 = simplify_bdd(temp1, c->conjunct[i]);
		bdd_free(gbm,temp1);
		temp1 = temp2;
		printf("->%ld",bdd_size(gbm,temp1,1));
	}
	printf("\n");

	return temp1;
}


/* Various tests for convergence. */

int conjunct_list_component_equality(conjunct_list a, conjunct_list b)
	/* This one does a component by component equality check. */
	/* Sufficient but not necessary. */
{
	int i;

	if (a->size != b->size) return FALSE;

	for (i=0; i < a->size; i++) {
		if (a->conjunct[i] != b->conjunct[i]) return FALSE;
	}

	return TRUE;
}

/* Helper functions for convergence_2 */

int check_disjunct_list_tautology(conjunct_list a)
	/* Returns TRUE or FALSE depending on whether the disjunct list a
		is a tautology. */
	/* Remember a is a DISJUNCT list! */
	/* This algorithm is based on my notes of 7/23/93 with later
	   additions. */
{
    conjunct_list x = copy_conjunct_list(a);
    int actual_size = x->size;
    int i,j;
    bdd temp1,temp2;
    bdd split_var;
    conjunct_list cofactor;

    /* Any disjunct is true makes whole disjunction true. */
    for (i=0; i < actual_size; i++) {
	if (x->conjunct[i] == bdd_one(gbm)) {
	    free_conjunct_list(x);
	    return TRUE;
	}
    }

    /* Any disjunct false can be discarded. */
    i=0;
    while (i < actual_size) {
	/* This loop pushes false to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (x->conjunct[i] == bdd_zero(gbm)) {
	    x->conjunct[i] = x->conjunct[--actual_size];
	    x->conjunct[actual_size] = bdd_zero(gbm);
	} else i++;
    }
    if (actual_size==0) {
	free_conjunct_list(x);
	return FALSE;
    }

    /* Any two disjuncts compliments makes whole disjunction true. */
    for (i=0; i < actual_size; i++) {
	temp1 = bdd_not(gbm,x->conjunct[i]);
	for (j=i+1; j < actual_size; j++) {
	    if (x->conjunct[j] == temp1) {
		bdd_free(gbm,temp1);	/* Memory leak spotted 8 Aug 94 AJH */
		free_conjunct_list(x);
		return TRUE;
	    }
	}
	bdd_free(gbm,temp1);
    }

    /* Any disjunct that duplicates another can be discarded. */
    i=0;
    while (i < actual_size) {
	/* Loop invariant:  conjuncts[i..actual_size-1] contains no copies of
	   anything in conjuncts[0..i-1]. */
	j=i+1;
	while (j < actual_size) {
	    /* This loop pushes duplicates to the end.  Termination guaranteed
	       since on each iteration j increases or actual_size decreases. */
	    if (x->conjunct[i]==x->conjunct[j]) {
		/* Duplicate */
		temp1 = x->conjunct[j];
		x->conjunct[j] = x->conjunct[--actual_size];
		x->conjunct[actual_size] = temp1;
	    } else j++;
	}

	i++;
    }
    if (actual_size==0) {
	free_conjunct_list(x);
	return FALSE;
    }

    /* Can replace any disjunct by the same disjunct simplified by any
	other disjunct.  If the simplification is David Long's bdd_reduce,
	this case also takes care of checking if any two disjuncts union
	to TRUE. */
    for (i=0; i < actual_size; i++) {
	temp1 = bdd_not(gbm,x->conjunct[i]);
	for (j=i; j < actual_size; j++) {
	    if (i!=j) {
		temp2 = simplify_bdd(x->conjunct[j],temp1);
		bdd_free(gbm,x->conjunct[j]);
		x->conjunct[j] = temp2;
	    }
	}
	bdd_free(gbm,temp1);
    }
    /* Eliminate any TRUEs and FALSEs that came out of simplification. */
    for (i=0; i < actual_size; i++) {
	if (x->conjunct[i] == bdd_one(gbm)) {
	    free_conjunct_list(x);
	    return TRUE;
	}
    }
    i=0;
    while (i < actual_size) {
	/* This loop pushes false to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (x->conjunct[i] == bdd_zero(gbm)) {
	    x->conjunct[i] = x->conjunct[--actual_size];
	    x->conjunct[actual_size] = bdd_zero(gbm);
	} else i++;
    }
    if (actual_size==0) {
	free_conjunct_list(x);
	return FALSE;
    }

    /* Other checks:  containment. */
    /* CONTINUE HERE */

    /* Last resort:  Shannon (Boole) expand and recurse. */

    /* Nicer status monitoring. */
    if (actual_size >= 10) {
	fprintf(stderr,"FOOBIE %d\n",actual_size); fflush(stderr);
    } else {
	fprintf(stderr,"%d",actual_size); fflush(stderr);
    }

    split_var = bdd_if(gbm,x->conjunct[0]);	/* Simple policy for starters.*/
    cofactor = new_conjunct_list(actual_size);

    /* Positive cofactor */
    for (i=0; i < actual_size; i++) {
	cofactor->conjunct[i] = bdd_compose(gbm,
					x->conjunct[i],
					split_var,
					bdd_one(gbm));
    }
    if (!check_disjunct_list_tautology(cofactor)) {
	free_conjunct_list(cofactor);
	free_conjunct_list(x);
	fprintf(stderr,"\b \b"); fflush(stderr);
	return FALSE;
    }

    fprintf(stderr,"."); fflush(stderr);

    /* Negative cofactor */
    for (i=0; i < actual_size; i++) {
	bdd_free(gbm,cofactor->conjunct[i]);
	cofactor->conjunct[i] = bdd_compose(gbm,
					x->conjunct[i],
					split_var,
					bdd_zero(gbm));
    }

    if (!check_disjunct_list_tautology(cofactor)) {
	free_conjunct_list(cofactor);
	free_conjunct_list(x);
	fprintf(stderr,"\b\b  \b\b"); fflush(stderr);
	return FALSE;
    }

    fprintf(stderr,"\b\b  \b\b"); fflush(stderr);

    free_conjunct_list(cofactor);
    free_conjunct_list(x);
    /*fprintf(stderr,"FOOBIE 7 actual_size=%d\n",actual_size);*/
    return TRUE;	/* Both cofactors are tautologies. */
}

int conjunct_list_implies_bdd(conjunct_list a, bdd b)
{
    conjunct_list x;
    int i;

    x = new_conjunct_list(a->size + 1);
    for (i=0; i < a->size; i++)
	x->conjunct[i] = bdd_not(gbm,a->conjunct[i]);
    x->conjunct[a->size] = bdd_identity(gbm,b);

    /* At this point, x is actually the implicit *disjunction* ((not a) or b)
       which is a tautology iff a implies b. */
    i = check_disjunct_list_tautology(x);
    free_conjunct_list(x);
    return i;
}

int check_conjunct_list_implication_step(conjunct_list x, conjunct_list y)
    /* Checks whether x implies y. */
    /* 8 August 1994, AJH -- I realized that the way I had been decomposing
	this problem has two weaknesses:  (1) it might be easy to disprove
	implication for one conjunct, but not the others, so it'd be nice
	to detect the easier conjuncts faster and (2) the current version
	ends up redoing the Shannon decomposition of the conjunct list over
	and over again.  So, this is an attempt at doing it more directly. */
    /* MAY MODIFY ARGUMENTS.  Use check_conjunct_list_implication() as
	a wrapper to protect arguments. 21 August 1994 AJH */
{
    int x_size = x->size;
    int y_size = y->size;
    int i,j;
    bdd temp1,temp2;
    bdd split_var;
    conjunct_list x_cofactor, y_cofactor;

    /* Any x[i] false makes whole implication true. */
    for (i=0; i < x_size; i++) {
	if (x->conjunct[i] == bdd_zero(gbm)) {
	    return TRUE;
	}
    }

    /* Any x[i] true can be discarded. */
    i=0;
    while (i < x_size) {
	/* This loop pushes true to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (x->conjunct[i] == bdd_one(gbm)) {
	    x->conjunct[i] = x->conjunct[--x_size];
	    x->conjunct[x_size] = bdd_one(gbm);
	} else i++;
    }

    /* Any y[i] false makes whole consequent false. */
    for (i=0; i < y_size; i++) {
	if (y->conjunct[i] == bdd_zero(gbm)) {
	    y->conjunct[i] = y->conjunct[0];
	    y->conjunct[0] = bdd_zero(gbm);
	    y_size = 1;
	    break;
	}
    }

    /* Any y[i] true can be discarded. */
    i=0;
    while (i < y_size) {
	/* This loop pushes true to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (y->conjunct[i] == bdd_one(gbm)) {
	    y->conjunct[i] = y->conjunct[--y_size];
	    y->conjunct[y_size] = bdd_one(gbm);
	} else i++;
    }

    /* Check for empty lists. */
    if (y_size==0) {
	/* Consequent is TRUE. */
	return TRUE;
    } else if (x_size==0) {
	/* Antecedent is TRUE, but consequent isn't. */
	return FALSE;
    }

    /* Any complements in antecedent make whole implication true. */
    for (i=0; i < x_size; i++) {
	temp1 = bdd_not(gbm,x->conjunct[i]);
	for (j=i+1; j < x_size; j++) {
	    if (x->conjunct[j] == temp1) {
		bdd_free(gbm,temp1);
		return TRUE;
	    }
	}
	bdd_free(gbm,temp1);
    }

    /* Any complements in consequent make whole consequent false. */
    for (i=0; i < y_size; i++) {
	int escape = 0;
	temp1 = bdd_not(gbm,y->conjunct[i]);
	for (j=i+1; j < y_size; j++) {
	    if (y->conjunct[j] == temp1) {
		bdd_free(gbm,temp1);
		bdd_free(gbm,y->conjunct[j]);
		y->conjunct[j] = y->conjunct[0];
		y->conjunct[0] = bdd_zero(gbm);
		y_size = 1;
		escape=1;
		break;
	    }
	}
	if (escape) break;
	bdd_free(gbm,temp1);
    }

    /* Complement between antecedent and consequent makes whole implication
	false, unless entire antecedent is false.  Too messy, not implemented.*/

    /* Any duplicates in antecedent can be discarded. */
    i=0;
    while (i < x_size) {
	/* Loop invariant:  conjuncts[i..x_size-1] contains no copies of
	   anything in conjuncts[0..i-1]. */
	j=i+1;
	while (j < x_size) {
	    /* This loop pushes duplicates to the end.  Termination guaranteed
	       since on each iteration j increases or x_size decreases. */
	    if (x->conjunct[i]==x->conjunct[j]) {
		/* Duplicate */
		temp1 = x->conjunct[j];
		x->conjunct[j] = x->conjunct[--x_size];
		x->conjunct[x_size] = temp1;
	    } else j++;
	}

	i++;
    }

    /* Any duplicates in consequent can be discarded. */
    i=0;
    while (i < y_size) {
	/* Loop invariant:  conjuncts[i..y_size-1] contains no copies of
	   anything in conjuncts[0..i-1]. */
	j=i+1;
	while (j < y_size) {
	    /* This loop pushes duplicates to the end.  Termination guaranteed
	       since on each iteration j increases or y_size decreases. */
	    if (y->conjunct[i]==y->conjunct[j]) {
		/* Duplicate */
		temp1 = y->conjunct[j];
		y->conjunct[j] = y->conjunct[--y_size];
		y->conjunct[y_size] = temp1;
	    } else j++;
	}

	i++;
    }

    /* Can simplify any x[i] by any other x[j].  Can simplify any y[i] by
	any other y[j] or x[j].  If the simplify is decent (like David Long's
	bdd_reduce), this should handle many disjoint-conjunction and
	implication special cases. */
    for (i=0; i < x_size; i++) {
	for (j=0; j < x_size; j++) {
	    if (i!=j) {
		temp2 = simplify_bdd(x->conjunct[j],x->conjunct[i]);
		bdd_free(gbm,x->conjunct[j]);
		x->conjunct[j] = temp2;
	    }
	}
    }
    for (i=0; i < y_size; i++) {
	for (j=0; j < y_size; j++) {
	    if (i!=j) {
		temp2 = simplify_bdd(y->conjunct[j],y->conjunct[i]);
		bdd_free(gbm,y->conjunct[j]);
		y->conjunct[j] = temp2;
	    }
	}
    }
    for (i=0; i < x_size; i++) {
	for (j=0; j < y_size; j++) {
	    temp2 = simplify_bdd(y->conjunct[j],x->conjunct[i]);
	    bdd_free(gbm,y->conjunct[j]);
	    y->conjunct[j] = temp2;
	}
    }
    /* Eliminate any TRUEs and FALSEs that came out of simplification. */
    /* Any x[i] false makes whole implication true. */
    for (i=0; i < x_size; i++) {
	if (x->conjunct[i] == bdd_zero(gbm)) {
	    return TRUE;
	}
    }

    /* Any x[i] true can be discarded. */
    i=0;
    while (i < x_size) {
	/* This loop pushes true to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (x->conjunct[i] == bdd_one(gbm)) {
	    x->conjunct[i] = x->conjunct[--x_size];
	    x->conjunct[x_size] = bdd_one(gbm);
	} else i++;
    }

    /* Any y[i] false makes whole consequent false. */
    for (i=0; i < y_size; i++) {
	if (y->conjunct[i] == bdd_zero(gbm)) {
	    y->conjunct[i] = y->conjunct[0];
	    y->conjunct[0] = bdd_zero(gbm);
	    y_size = 1;
	    break;
	}
    }

    /* Any y[i] true can be discarded. */
    i=0;
    while (i < y_size) {
	/* This loop pushes true to end of list.  Termination guaranteed
	   because on each iteration i increases or actual_size decreases. */
	if (y->conjunct[i] == bdd_one(gbm)) {
	    y->conjunct[i] = y->conjunct[--y_size];
	    y->conjunct[y_size] = bdd_one(gbm);
	} else i++;
    }

    /* Check for empty lists. */
    if (y_size==0) {
	/* Consequent is TRUE. */
	return TRUE;
    } else if (x_size==0) {
	/* Antecedent is TRUE, but consequent isn't. */
	return FALSE;
    }

    /* Other checks:  containment. */
    /* CONTINUE HERE */

    /* Last resort:  Shannon (Boole) expand and recurse. */

    /* Status monitoring. */
    /*fprintf(stderr,".");*/

    split_var = bdd_if(gbm,x->conjunct[0]);	/* Simple policy for starters.*/
	/* Note that at this point, x->conjunct[0] cannot be just TRUE
		or else x_size would have been zero and we wouldn't be here,
		nor could it be just FALSE, as we would have caught that. */

    x_cofactor = new_conjunct_list(x_size);
    y_cofactor = new_conjunct_list(y_size);

    /* Positive cofactor */
    for (i=0; i < x_size; i++) {
	x_cofactor->conjunct[i] = bdd_compose(gbm,
					x->conjunct[i],
					split_var,
					bdd_one(gbm));
    }
    for (i=0; i < y_size; i++) {
	y_cofactor->conjunct[i] = bdd_compose(gbm,
					y->conjunct[i],
					split_var,
					bdd_one(gbm));
    }
    if (!check_conjunct_list_implication_step(x_cofactor,y_cofactor)) {
	free_conjunct_list(x_cofactor);
	free_conjunct_list(y_cofactor);
	/*fprintf(stderr,"\b \b"); fflush(stderr);*/
	return FALSE;
    }

    /* Negative cofactor */
    for (i=0; i < x_size; i++) {
	bdd_free(gbm,x_cofactor->conjunct[i]);
	x_cofactor->conjunct[i] = bdd_compose(gbm,
					x->conjunct[i],
					split_var,
					bdd_zero(gbm));
    }
    for (i=0; i < y_size; i++) {
	bdd_free(gbm,y_cofactor->conjunct[i]);
	y_cofactor->conjunct[i] = bdd_compose(gbm,
					y->conjunct[i],
					split_var,
					bdd_zero(gbm));
    }
    if (!check_conjunct_list_implication_step(x_cofactor,y_cofactor)) {
	free_conjunct_list(x_cofactor);
	free_conjunct_list(y_cofactor);
	/*fprintf(stderr,"\b \b"); fflush(stderr);*/
	return FALSE;
    }

    /*fprintf(stderr,"\b \b"); fflush(stderr);*/

    free_conjunct_list(x_cofactor);
    free_conjunct_list(y_cofactor);
    return TRUE;	/* Implication holds for both cofactors. */
}	/* check_conjunct_list_implication_step */

int check_conjunct_list_implication(conjunct_list a, conjunct_list b)
    /* Checks whether a implies b. */
    /* Wrapper to reduce amount of memory alloc/dealloc. */
{
	int result;
	conjunct_list x, y;

	x = copy_conjunct_list(a);
	y = copy_conjunct_list(b);
	result = check_conjunct_list_implication_step(x,y);
	free_conjunct_list(x);
	free_conjunct_list(y);
	return result;
}

int check_disjunct_list_implication(disjunct_list a, disjunct_list b)
    /* Checks whether a implies b. */
    /* a implies b iff ~b implies ~a, and then we can use de Morgan's laws
	to get conjunct lists. */
{
	conjunct_list not_a, not_b;
	int i;
	int result;

	not_a = de_morgan_dl_to_cl(a);
	not_b = de_morgan_dl_to_cl(b);

	result = check_conjunct_list_implication_step(not_b, not_a);

	free_conjunct_list(not_a);
	free_conjunct_list(not_b);

	return result;
}

int check_conjunct_list_equality(conjunct_list a, conjunct_list b)
	/* This one is necessary and sufficient.  It works by checking
		implication in both directions. */
{
    /* Check that a implies b. */
    if (!check_conjunct_list_implication(a,b)) return FALSE;
    if (!check_conjunct_list_implication(b,a)) return FALSE;
    return TRUE;
}


int check_disjunct_list_implies_conjunct_list(disjunct_list a, conjunct_list b)
	/* True iff each disjunct in a implies all conjuncts in b. */
{
    int i, j;
    bdd temp;

    for (i=0; i < a->size; i++) {
	for (j=0; j < b->size; j++) {
	    /* David Long changed his bdd_implies function so that it's
	       weird now.  It returns a bdd that implies (f and not g).
	       Therefore, f implies g iff bdd_implies(f,g) returns bdd_zero. */
	    temp = bdd_implies(gbm,a->disjunct[i],b->conjunct[j]);
	    if (temp != bdd_zero(gbm)) {
		bdd_free(gbm,temp);
		return FALSE;
	    }
	    bdd_free(gbm,temp);
	}
    }

    return TRUE;
}


/* This is for benchmarking data only.  Should be commented out. */
void print_conjunct_list_size(conjunct_list x, int d)
{
	int i;
	bdd *foo = (bdd *)malloc((x->size+1)*sizeof(bdd));

	printf("Iteration %d:\n",d);
	for (i=0; i < x->size; i++) {
		foo[i] = x->conjunct[i];
	}
	foo[i] = NULL;

	for (i=0; i < x->size; i++) {
		printf("\t%d: %ld\n",i,bdd_size(gbm,foo[i],1));
		/*bdd_print_profile(gbm,foo[i],bdd_naming_fn_full,NULL,80,stdout);*/
	}
	printf("Total Size: %ld\n",bdd_size_multiple(gbm,foo,1));

	free(foo);
}

void print_disjunct_list_size(disjunct_list x, int d)
{
	int i;
	bdd *foo = (bdd *)malloc((x->size+1)*sizeof(bdd));

	printf("Iteration %d:\n",d);
	for (i=0; i < x->size; i++) {
		foo[i] = x->disjunct[i];
	}
	foo[i] = NULL;

	for (i=0; i < x->size; i++) {
		printf("\t%d: %ld\n",i,bdd_size(gbm,foo[i],1));
		/*bdd_print_profile(gbm,foo[i],bdd_naming_fn_full,NULL,80,stdout);*/
	}
	printf("Total Size: %ld\n",bdd_size_multiple(gbm,foo,1));

	free(foo);
}

conjunct_list bdd_to_conjunct_list(bdd x)
{
    conjunct_list y;
    y = new_conjunct_list(1);
    y->conjunct[0] = bdd_identity(gbm,x);
    return y;
}

bdd conjunct_list_to_bdd(conjunct_list x)
{
    bdd y, temp1;
    int i;

    y = bdd_one(gbm);
    for (i=0; i < x->size; i++) {
	temp1 = bdd_and(gbm,y,x->conjunct[i]);
	bdd_free(gbm,y);
	y = temp1;
    }

    return y;
}
