/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Variable Ordering Routines Version 0.0 */
/* By Alan J. Hu */
/* November 8, 1991 */

/* ************************ CHANGE LOG **************************** */
/*
24 August 1992	AJH -- Added flip routine to put integers (bit-vector types)
			MSB-first in the variable order.
31 August 1992	AJH -- Disabled flip routine, since it's interacting
			incorrectly with interleaving.
*/
/* **************************************************************** */

#include "ever.h"
#include "y.tab.h"	/* Defines constants for tokens. */


void randomize_variables(bdd *base, type_desc remaining_type)
{
	/* Randomly reorders variables.  This was a debugging stub,
	   but I thought it might be nice to keep around.  */
	long random();

	int i, j;
	bdd temp;

	srandom((long)getpid());
	for (i=0; i < remaining_type->bdd_count; i++)
	  for (j=i+1; j < remaining_type->bdd_count; j++)
	    if (random() & 01) {
		temp = base[i];
		base[i] = base[j];
		base[j] = temp;
	    }
}



void flip_bits(bdd *base, type_desc bits)
	/* No longer used.  Is incorrect with interleaving. AJH 8/31/92 */
	/* Helper function for interleave_variables(), although this
		is conceptually different from interleaving.  Flips
		the ordering of the bits. */
{
	int i, j;
	bdd temp0, temp1, temp2;

	for (i=0, j=bits->bdd_count/3 - 1; i<j; i++, j--) {
		/* Flip the bdd vars from both ends. */
fprintf(stderr,"DEBUG--Flipping %d\n",i);
		temp0 = base[3*i+0];
		temp1 = base[3*i+1];
		temp2 = base[3*i+2];
		base[3*i+0] = base[3*j+0];
		base[3*i+1] = base[3*j+1];
		base[3*i+2] = base[3*j+2];
		base[3*j+0] = temp0;
		base[3*j+1] = temp1;
		base[3*j+2] = temp2;
	}
}

void interleave_array(bdd *base, type_desc array)
	/* Helper function for interleave_variables().  Assumes that
	   the elements have been interleaved already (if they were
	   supposed to be).  This routine does the actual pointer
	   moving at this level of the recursion. */
{
    bdd *temp;
    int element_size = array->subtype->bdd_count / 3;
    int element_count = array->array_ub - array->array_lb + 1;
    int which_element, which_bit;
    int i;

    /* Make a copy of the current variable ordering. */
    temp = (bdd *)calloc(array->bdd_count,sizeof(bdd));
    bcopy((char *)base,(char *)temp,array->bdd_count*sizeof(bdd));

    /* Interleave the variables! */
    for (i=0; i < element_size*element_count; i++) {
	/* Determine which bit of which array element i points to. */
	which_element = i / element_size;
	which_bit = i % element_size;
	/* Pull out the appropriate variable from the ordering.
	   All element_count low-order bits come first, etc. */
	base[3*i] = temp[3*(which_bit*element_count+which_element)];
	base[3*i+1] = temp[3*(which_bit*element_count+which_element)+1];
	base[3*i+2] = temp[3*(which_bit*element_count+which_element)+2];
    }

    free(temp);
}

void interleave_record(bdd *base, type_desc record)
	/* Helper function for interleave_variables().  Assumes that
	   the fields have been interleaved already (if they were
	   supposed to be).  This routine does the actual pointer
	   moving at this level of the recursion. */
{
    struct offsets {
			bdd *start;
			bdd *stop;
			struct offsets *next;
		    } *offset_head, *offset_scratch, *extra;
    bdd *temp;
    record_list field;
    int i;
    bdd *offset;

    /* Make a copy of the current variable ordering. */
    temp = (bdd *)calloc(record->bdd_count,sizeof(bdd));
    bcopy((char *)base,(char *)temp,record->bdd_count*sizeof(bdd));

    /* Make a list of offsets for each field. */
    field = record->field_list;
    offset = base;
    offset_scratch = NULL;
    while (field!=NULL) {
	/* offset_scratch always points to the last element */
	if (offset_scratch==NULL) {
	    offset_scratch = (struct offsets *)malloc(sizeof(struct offsets));
	    offset_head = offset_scratch;
	} else {
	    offset_scratch->next =
	    		(struct offsets *)malloc(sizeof(struct offsets));
	    offset_scratch = offset_scratch->next;
	}

	offset_scratch->start = offset;
	offset += field->field_type->bdd_count;
	offset_scratch->stop = offset;
	field = field->next;
    }
    if (offset_scratch!=NULL)	/* protect against empty records */
	offset_scratch->next = offset_head;	/* circularize list */
    extra = offset_scratch;	/* remember last item for freeing later. */
    offset_scratch = offset_head;

    /* Interleave the variables! */
    for (i=0; i < record->bdd_count/3; i++) {
	/* Step to the next field that still needs bits. */
	while (offset_scratch->start>=offset_scratch->stop)
	    offset_scratch = offset_scratch->next;
	
	/* Allocate a BDD variable for this field. */
	*(offset_scratch->start) = temp[3*i];
	offset_scratch->start++;
	*(offset_scratch->start) = temp[3*i+1];
	offset_scratch->start++;
	*(offset_scratch->start) = temp[3*i+2];
	offset_scratch->start++;

	/* Go to next field. */
	offset_scratch = offset_scratch->next;
    }

    /* Dispose of scratch memory. */
    if (extra!=NULL) extra->next=NULL;	/* De-circularize the list. */
    while (offset_head!=NULL) {
	offset_scratch = offset_head->next;
	free(offset_head);
	offset_head = offset_scratch;
    }
    free(temp);
}

void interleave_variables(bdd *base, type_desc remaining_type)
	/* Interleaves the BDD variables in the array pointed to by
	   base, using the interleavings specified in the type descriptor. */
{
    if (remaining_type->type==BITS) return;	/* Nothing to do. */

    if (remaining_type->type==ARRAY) {
	/* First interleave the entries. */
	int stepsize = remaining_type->subtype->bdd_count;
	int i;
	bdd *temp;

	for (i=remaining_type->array_lb, temp=base;
	     i <= remaining_type->array_ub;
	     i++, temp += stepsize) {
	    interleave_variables(temp, remaining_type->subtype);
	}

	/* Now, interleave the array if requested. */
	if (remaining_type->interleaved)
	    interleave_array(base, remaining_type);

	return;
    }

    if (remaining_type->type==RECORD) {
	/* First interleave the fields. */
	record_list fields = remaining_type->field_list;
	bdd *temp = base;

	while (fields!=NULL) {
	    interleave_variables(temp, fields->field_type);
	    temp += fields->field_type->bdd_count;
	    fields = fields->next;
	}

	/* Now, interleave the record if requested. */
	if (remaining_type->interleaved)
	    interleave_record(base, remaining_type);

	return;
    }
}

